#! /usr/bin/env python
###############################################################################
# Copyright (C) 2015
# ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
###############################################################################

# Usage:
# . python tc_apertif_fb_bf_xc.py --gn 0:32 --sp 0:3
# Purpose:
# . Use a beamformer rack to generate 24 signals and correlate them in 1 XC node
# Description:
# . The 24 generated signals represent 12 telescopes*2 pols.
# . The 24 signals are all correlated in apertif_unb1_correlator and will yield a phase plot for 1 channel.
#   . So these will yield 300 visibilities, even in the '48 visiblity correlator' (although only 1 channel).
# . The following waveforms are generated in the back nodes. The target output shows where
#   the signal should be routed to (unmodified by weights, e.g. weight=1)
#   -----+----+-----+-----------+-----------+-------------+-------------------------------------+------------------------
#   UNB  | BN | SP  | Frequency | Amplitude | Phase shift | Target output                       | To correlator 10G input
#   -----+----+-----+-----------+-----------+-------------+-------------------------------------+------------------------
#     0  |  0 |  0  | 80 MHz    | 0.5       |  0 deg      | UNB3,FN0|beamlet 0 (out of 0..1023) | UNB4, FN0, RX0
#        |    |  1  | 80 MHz    | 0.5       |  1 deg      | UNB3,FN0|beamlet 1                  | UNB4, FN0, RX0
#        |    |  2  | 80 MHz    | 0.5       |  2 deg      | UNB3,FN0|beamlet 2                  | UNB4, FN0, RX0
#        |    |  3  | 80 MHz    | 0.5       |  3 deg      | UNB3,FN0|beamlet 3                  | UNB4, FN0, RX0 
#        |  1 |  0  | 80 MHz    | 0.5       |  4 deg      | UNB3,FN0|beamlet 4                  | UNB4, FN0, RX0
#        |    |  1  | 80 MHz    | 0.5       |  5 deg      | UNB3,FN0|beamlet 5                  | UNB4, FN0, RX0 
#        |    |  2  | 80 MHz    | 0.5       |  6 deg      | UNB3,FN0|beamlet 6                  | UNB4, FN0, RX0
#        |    |  3  | 80 MHz    | 0.5       |  7 deg      | UNB3,FN0|beamlet 7                  | UNB4, FN0, RX0
#        |  2 |  0  | 80 MHz    | 0.5       |  8 deg      | UNB3,FN1|beamlet 0                  | UNB4, FN0, RX1
#        |    |  1  | 80 MHz    | 0.5       |  9 deg      | UNB3,FN1|beamlet 1                  | UNB4, FN0, RX1
#        |    |  2  | 80 MHz    | 0.5       | 10 deg      | UNB3,FN1|beamlet 2                  | UNB4, FN0, RX1
#        |    |  3  | 80 MHz    | 0.5       | 11 deg      | UNB3,FN1|beamlet 3                  | UNB4, FN0, RX1
#        |  3 |  0  | 80 MHz    | 0.5       | 12 deg      | UNB3,FN1|beamlet 4                  | UNB4, FN0, RX1
#        |    |  1  | 80 MHz    | 0.5       | 13 deg      | UNB3,FN1|beamlet 5                  | UNB4, FN0, RX1
#        |    |  2  | 80 MHz    | 0.5       | 14 deg      | UNB3,FN1|beamlet 6                  | UNB4, FN0, RX1
#        |    |  3  | 80 MHz    | 0.5       | 15 deg      | UNB3,FN1|beamlet 7                  | UNB4, FN0, RX1
#     1  |  0 |  0  | 80 MHz    | 0.5       | 16 deg      | UNB3,FN2|beamlet 0                  | UNB4, FN0, RX2
#        |    |  1  | 80 MHz    | 0.5       | 17 deg      | UNB3,FN2|beamlet 1                  | UNB4, FN0, RX2
#        |    |  2  | 80 MHz    | 0.5       | 18 deg      | UNB3,FN2|beamlet 2                  | UNB4, FN0, RX2
#        |    |  3  | 80 MHz    | 0.5       | 19 deg      | UNB3,FN2|beamlet 3                  | UNB4, FN0, RX2
#        |  1 |  0  | 80 MHz    | 0.5       | 20 deg      | UNB3,FN2|beamlet 4                  | UNB4, FN0, RX2
#        |    |  1  | 80 MHz    | 0.5       | 21 deg      | UNB3,FN2|beamlet 5                  | UNB4, FN0, RX2
#        |    |  2  | 80 MHz    | 0.5       | 22 deg      | UNB3,FN2|beamlet 6                  | UNB4, FN0, RX2
#        |    |  3  | 80 MHz    | 0.5       | 23 deg      | UNB3,FN2|beamlet 7                  | UNB4, FN0, RX2
#

# 80MHz * 0.781250 = 102.4 so almost in the center of subband 102.
FREQUENCY = 0.1  # out of range 0.0 .. 1.0 (1.0 = 800MHz)
AMPLITUDE = 0.9  # our of range 0.0 .. 1.0

NOF_UNB_PER_RACK = 4
NOF_BN_PER_UNB   = 4
NOF_SP_PER_BN    = 4

###############################################################################
# Standard IO, TC creation
###############################################################################
import test_case
import node_io

tc = test_case.Testcase('TC - ', '')
tc.set_result('PASSED')
io = node_io.NodeIO(tc.nodeImages, tc.base_ip)

###############################################################################
# Stop data path
###############################################################################
import pi_bsn_source
bsSrc = pi_bsn_source.PiBsnSource(tc, io, tc.nodeBnNrs)

# Disable      
bsSrc.write_disable()    

###############################################################################
# Generate 24 phase-shifted sine waves in the back nodes
###############################################################################
import pi_diag_wg_wideband

# Create 64 waveform generators: 4UNB*4BN*4SP.
wg = []
for bi in range(NOF_UNB_PER_RACK*NOF_BN_PER_UNB):
    for sp in range(NOF_SP_PER_BN):  
        wg.append(pi_diag_wg_wideband.PiDiagWgWideband(tc, io, sp, tc.nodeBnNrs[bi]))

# Disable all WGs
for i in range(64):
    wg[i].write_mode_off()

# Create 24 phase shifted sine waves on WG 0..23
for sp in range(24):
    phase = float(sp)
    wg[sp].write_sinus_settings(phase, FREQUENCY, AMPLITUDE)

# We don't want WG 24..63 to output anything
for sp in range(24,64):
    wg[sp].write_sinus_settings(0.0, 0.0, 0.0)

# Enable all WGs
for i in range(64):
    wg[i].write_mode_sinus()

###############################################################################
# Start data path
###############################################################################
import pi_ppsh
ppsh  = pi_ppsh.PiPpsh(tc, io, tc.nodeBnNrs)

# wait until PPS
ppsh.wait_until_pps()  

# (re)start all at next PPS       
bsSrc.write_restart_pps()    

# wait until PPS
ppsh.wait_until_pps()  

# wait until PPS
ppsh.wait_until_pps()  

###############################################################################
# Plot subband statistics to show our spike
# . Why does our plot look different from
#   python $UPE/apps/apertif_beamformer/util_apertif_beamformer.py --unb 0:3 --bn 0:3 --fn 0:3 --sp 0:3 -v 5 --cmd 78
# . If we choose bin numbers on the X axis, why is our plot inverted?
#   . Freq = 640MHz/800MHz (320MHz/400MHz) which is bin (0.781250*320MHz=)250, right?
#     . Then why does our signal show up in bin 204?
###############################################################################
import Stati_functions
import unb_apertif as apr
import pi_st_sst

sst = []
for i in xrange(apr.c_nof_bf_units*apr.c_nof_wb_ffts):                                                                                       
    sst.append( pi_st_sst.PiStSst(tc, io, apr.c_fft_size/apr.c_wb_factor, apr.c_stat_data_w, apr.c_stat_data_sz, False, i, tc.nodeBnNrs))

c_blocks_per_sync  = 781250
c_stat_input_bits  = 16     #   16    Number of bits that go into the statistics module. Used to determine the normalization factor.
xas_type='freq_z2' #'freq_z2'  # freq_z2 for normal freq axis, freq for 0-fs/2, bin for bin numbers  
xas_type='bin' # werkt niet
fig_cnt=1

stati_conf = Stati_functions.statiConfiguration(apr.c_fft_size, apr.c_nof_wb_ffts, apr.c_bsn_period, c_blocks_per_sync,\
                                                c_stat_input_bits, apr.c_wb_factor, apr.c_nof_bf_units)

stati_h = Stati_functions.Stati_functions(stati_conf)
stati_h.plot_init();
stati_h.subband_stati_capture(tc, sst) 
stati_h.plot_subband_stati(tc, figs='one', xas=xas_type, norm='dB',lim_axis='off', bck_fft='off', fig_nr=fig_cnt,pl_legend='off')

close_figures = 0
if(close_figures == 1):
    stati_h.close_figs(tc)
else:
    stati_h.plot_last(tc)

###############################################################################
# Select subband 102
###############################################################################




