-------------------------------------------------------------------------------
--
-- Copyright (C) 2012
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- JIVE (Joint Institute for VLBI in Europe) <http://www.jive.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

LIBRARY IEEE, common_lib, unb1_board_lib, eth_lib, diag_lib, dp_lib, ddr3_lib, tech_ddr_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE IEEE.NUMERIC_STD.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE unb1_board_lib.unb1_board_pkg.ALL;
USE eth_lib.eth_pkg.ALL;
USE diag_lib.diag_pkg.ALL; 
USE dp_lib.dp_stream_pkg.ALL;
USE tech_ddr_lib.tech_ddr_pkg.ALL;
USE ddr3_lib.ddr3_pkg.ALL; 

ENTITY unb1_ddr3_transpose IS
  GENERIC (
    g_sim         : BOOLEAN      := FALSE; --Overridden by TB
    g_sim_unb_nr  : NATURAL      := 0;
    g_sim_node_nr : NATURAL      := 0;
    g_design_name : STRING       := "unb1_ddr3_transpose";
    g_design_note : STRING       := "Test Design";
    --g_technology  : NATURAL      := c_tech_stratixiv;
    g_stamp_date  : NATURAL      := 0;  -- Date (YYYYMMDD) -- set by QSF
    g_stamp_time  : NATURAL      := 0;  -- Time (HHMMSS)   -- set by QSF
    g_stamp_svn   : NATURAL      := 0;  -- SVN revision    -- set by QSF
    g_use_MB_I    : NATURAL      := 1   -- 1: use MB_I  0: do not use
  );
  PORT (
    -- GENERAL
    CLK          : IN    STD_LOGIC; -- System Clock
    PPS          : IN    STD_LOGIC; -- System Sync
    WDI          : OUT   STD_LOGIC; -- Watchdog Clear
    INTA         : INOUT STD_LOGIC; -- FPGA interconnect line
    INTB         : INOUT STD_LOGIC; -- FPGA interconnect line

    -- Others
    VERSION      : IN    STD_LOGIC_VECTOR(c_unb1_board_aux.version_w-1 DOWNTO 0);
    ID           : IN    STD_LOGIC_VECTOR(c_unb1_board_aux.id_w-1 DOWNTO 0);
    TESTIO       : INOUT STD_LOGIC_VECTOR(c_unb1_board_aux.testio_w-1 DOWNTO 0);
    
    -- I2C Interface to Sensors
    sens_sc      : INOUT STD_LOGIC;
    sens_sd      : INOUT STD_LOGIC;
  
    -- 1GbE Control Interface
    ETH_clk      : IN    STD_LOGIC;
    ETH_SGIN     : IN    STD_LOGIC;
    ETH_SGOUT    : OUT   STD_LOGIC;
    
    -- SO-DIMM Memory Bank I
    MB_I_IN       : IN    t_tech_ddr3_phy_in := c_tech_ddr3_phy_in_x;
    MB_I_IO       : INOUT t_tech_ddr3_phy_io;
    MB_I_OU       : OUT   t_tech_ddr3_phy_ou
  );
END unb1_ddr3_transpose;


ARCHITECTURE str OF unb1_ddr3_transpose IS

  -- Firmware version x.y
  CONSTANT c_fw_version             : t_unb1_board_fw_version := (0, 9);
  -- In simulation we don't need the 1GbE core for MM control, deselect it in c_use_phy based on g_sim
  CONSTANT c_use_phy                : t_c_unb1_board_use_phy  := (sel_a_b(g_sim, 0, 1), 0, 0, 0, 1, 0, 0, 1);
  
  -- Compose the Constants for the DUT   
  CONSTANT c_wr_chunksize            : POSITIVE   := sel_a_b(g_sim, 64, 64);-- 256);
  CONSTANT c_wr_nof_chunks           : POSITIVE   := sel_a_b(g_sim,  1,  1);--   1);
  CONSTANT c_rd_chunksize            : POSITIVE   := sel_a_b(g_sim, 16, 16);--  16);
  CONSTANT c_rd_nof_chunks           : POSITIVE   := sel_a_b(g_sim,  4,  4);--  16);
  CONSTANT c_gapsize                 : NATURAL    := sel_a_b(g_sim,  0,  0);--   0);
  CONSTANT c_nof_blocks              : POSITIVE   := sel_a_b(g_sim,  4,  4);--  16);
 
  CONSTANT c_ddr3_seq_conf           : t_ddr3_seq := (c_wr_chunksize, 
                                                      c_wr_nof_chunks,
                                                      c_rd_chunksize, 
                                                      c_rd_nof_chunks,
                                                      c_gapsize,      
                                                      c_nof_blocks);   
  
  CONSTANT c_blocksize               : POSITIVE := c_wr_nof_chunks * c_wr_chunksize;  

  CONSTANT c_ddr                     : t_c_ddr3_phy := c_ddr3_phy_4g;
  CONSTANT c_mts                     : NATURAL  := 800;--1066; --800
  CONSTANT c_phy                     : NATURAL  := 1;
  CONSTANT c_data_w                  : NATURAL  := 8; 
  CONSTANT c_nof_streams             : POSITIVE := 4;  
  CONSTANT c_frame_size_in           : POSITIVE := c_wr_chunksize;
  CONSTANT c_frame_size_out          : POSITIVE := c_frame_size_in;      
  CONSTANT c_nof_blk_per_sync        : POSITIVE := sel_a_b(g_sim, 16, 1600000);
  CONSTANT c_ena_pre_transpose       : BOOLEAN  := FALSE; 
 
  -- Custom definitions of constants
  CONSTANT c_bg_block_len           : NATURAL  := c_blocksize * c_rd_chunksize;
  CONSTANT c_db_block_len           : NATURAL  := c_blocksize * c_rd_chunksize;
 
  -- Configuration of the block generator:
  CONSTANT c_bg_buf_dat_w           : POSITIVE := c_nof_complex*c_data_w;
  CONSTANT c_bg_buf_adr_w           : POSITIVE := ceil_log2(c_bg_block_len);
  CONSTANT c_bg_data_file_prefix    : STRING   := "UNUSED"; -- "../../../src/hex/tb_bg_dat";
  CONSTANT c_bg_data_file_index_arr : t_nat_natural_arr := array_init(0, 128, 1);
  
  -- Configuration of the databuffers:
  CONSTANT c_db_data_w              : POSITIVE := c_diag_db_max_data_w;
  CONSTANT c_db_buf_nof_data        : POSITIVE := c_db_block_len;
  CONSTANT c_db_buf_use_sync        : BOOLEAN  := TRUE;
  CONSTANT c_db_data_type_re        : t_diag_data_type_enum := e_real;
  CONSTANT c_db_data_type_im        : t_diag_data_type_enum := e_imag;

  -- System
  SIGNAL cs_sim                     : STD_LOGIC;
  SIGNAL xo_clk                     : STD_LOGIC;
  SIGNAL xo_rst                     : STD_LOGIC;
  SIGNAL xo_rst_n                   : STD_LOGIC;
  SIGNAL mm_clk                     : STD_LOGIC;
  SIGNAL mm_locked                  : STD_LOGIC;
  SIGNAL mm_rst                     : STD_LOGIC;
  
  SIGNAL dp_rst                     : STD_LOGIC;
  SIGNAL dp_clk                     : STD_LOGIC;   
  SIGNAL dp_pps                     : STD_LOGIC;
  
  
  -- PIOs
  SIGNAL pout_wdi                   : STD_LOGIC;

  -- WDI override
  SIGNAL reg_wdi_mosi               : t_mem_mosi;
  SIGNAL reg_wdi_miso               : t_mem_miso;

  -- PPSH
  SIGNAL reg_ppsh_mosi              : t_mem_mosi;
  SIGNAL reg_ppsh_miso              : t_mem_miso;
  
  -- UniBoard system info
  SIGNAL reg_unb_system_info_mosi   : t_mem_mosi;
  SIGNAL reg_unb_system_info_miso   : t_mem_miso;
  SIGNAL rom_unb_system_info_mosi   : t_mem_mosi;
  SIGNAL rom_unb_system_info_miso   : t_mem_miso;

  -- UniBoard I2C sens
  SIGNAL reg_unb_sens_mosi          : t_mem_mosi;
  SIGNAL reg_unb_sens_miso          : t_mem_miso;

  -- eth1g
  SIGNAL eth1g_tse_clk              : STD_LOGIC;
  SIGNAL eth1g_mm_rst               : STD_LOGIC;
  SIGNAL eth1g_tse_mosi             : t_mem_mosi;  -- ETH TSE MAC registers
  SIGNAL eth1g_tse_miso             : t_mem_miso;
  SIGNAL eth1g_reg_mosi             : t_mem_mosi;  -- ETH control and status registers
  SIGNAL eth1g_reg_miso             : t_mem_miso;
  SIGNAL eth1g_reg_interrupt        : STD_LOGIC;   -- Interrupt
  SIGNAL eth1g_ram_mosi             : t_mem_mosi;  -- ETH rx frame and tx frame memory
  SIGNAL eth1g_ram_miso             : t_mem_miso;
  --SIGNAL eth1g_led                  : t_tse_led;

  -- Blockgenerator
  SIGNAL bg_siso_arr                : t_dp_siso_arr(c_nof_streams -1 DOWNTO 0) := (OTHERS => c_dp_siso_rdy);
  SIGNAL bg_sosi_arr                : t_dp_sosi_arr(c_nof_streams -1 DOWNTO 0);
  
  SIGNAL reg_diag_bg_mosi           : t_mem_mosi;
  SIGNAL reg_diag_bg_miso           : t_mem_miso;
                                    
  SIGNAL ram_diag_bg_mosi           : t_mem_mosi;
  SIGNAL ram_diag_bg_miso           : t_mem_miso;
                                    
  -- Databuffer                               
  SIGNAL ram_diag_data_buf_re_mosi  : t_mem_mosi;
  SIGNAL ram_diag_data_buf_re_miso  : t_mem_miso;
                                    
  SIGNAL reg_diag_data_buf_re_mosi  : t_mem_mosi;
  SIGNAL reg_diag_data_buf_re_miso  : t_mem_miso;
                                    
  SIGNAL ram_diag_data_buf_im_mosi  : t_mem_mosi;
  SIGNAL ram_diag_data_buf_im_miso  : t_mem_miso;
                                    
  SIGNAL reg_diag_data_buf_im_mosi  : t_mem_mosi;
  SIGNAL reg_diag_data_buf_im_miso  : t_mem_miso;
  
  -- DDR3 Transpose
  SIGNAL ram_ss_ss_transp_mosi      : t_mem_mosi;
  SIGNAL ram_ss_ss_transp_miso      : t_mem_miso;  
  
  -- BNS Monitor
  SIGNAL reg_bsn_monitor_mosi       : t_mem_mosi; 
  SIGNAL reg_bsn_monitor_miso       : t_mem_miso; 
  
  SIGNAL reg_io_ddr_mosi            : t_mem_mosi; 
  SIGNAL reg_io_ddr_miso            : t_mem_miso; 
   
  SIGNAL bsn_sosi_arr               : t_dp_sosi_arr(1 DOWNTO 0) := (OTHERS => c_dp_sosi_rst);
  
  SIGNAL out_sosi_arr               : t_dp_sosi_arr(c_nof_streams-1 DOWNTO 0) := (OTHERS => c_dp_sosi_rst);
  SIGNAL out_siso_arr               : t_dp_siso_arr(c_nof_streams-1 DOWNTO 0) := (OTHERS => c_dp_siso_rdy);

  SIGNAL ddr_ref_rst                : STD_LOGIC;      

BEGIN

  -----------------------------------------------------------------------------
  -- General control function
  -----------------------------------------------------------------------------
  u_ctrl : ENTITY unb1_board_lib.ctrl_unb1_board
  GENERIC MAP (
    g_sim         => g_sim,
    g_design_name => g_design_name,
    g_stamp_date  => g_stamp_date,
    g_stamp_time  => g_stamp_time, 
    g_stamp_svn   => g_stamp_svn, 
    g_fw_version  => c_fw_version,
    g_mm_clk_freq => c_unb1_board_mm_clk_freq_50M,
    g_use_phy     => c_use_phy,
    g_dp_clk_use_pll => FALSE,
    g_aux         => c_unb1_board_aux
  )
  PORT MAP (
    -- Clock an reset signals
    cs_sim                   => cs_sim,
    xo_clk                   => xo_clk,
    xo_rst                   => xo_rst,
    xo_rst_n                 => xo_rst_n,

    mm_clk                   => mm_clk,
    mm_locked                => mm_locked,
    mm_rst                   => mm_rst,
    
    dp_rst                   => OPEN,--dp_rst,
    dp_clk                   => OPEN,--dp_clk,
    dp_pps                   => dp_pps,
    dp_rst_in                => dp_rst,
    dp_clk_in                => dp_clk,
    
    -- Toggle WDI
    pout_wdi                 => pout_wdi,

    -- MM buses
    -- . Manual WDI override
    reg_wdi_mosi             => reg_wdi_mosi,
    reg_wdi_miso             => reg_wdi_miso,
    -- . system_info
    reg_unb_system_info_mosi => reg_unb_system_info_mosi,
    reg_unb_system_info_miso => reg_unb_system_info_miso, 
    rom_unb_system_info_mosi => rom_unb_system_info_mosi,
    rom_unb_system_info_miso => rom_unb_system_info_miso, 
    -- . UniBoard I2C sensors
    reg_unb_sens_mosi        => reg_unb_sens_mosi,
    reg_unb_sens_miso        => reg_unb_sens_miso,    
    -- . PPSH
    reg_ppsh_mosi            => reg_ppsh_mosi,
    reg_ppsh_miso            => reg_ppsh_miso,
    
    -- eth1g
    eth1g_tse_clk            => eth1g_tse_clk,   -- 125 MHz from xo_clk PLL in SOPC system
    eth1g_mm_rst             => eth1g_mm_rst,
    eth1g_tse_mosi           => eth1g_tse_mosi,
    eth1g_tse_miso           => eth1g_tse_miso,
    eth1g_reg_mosi           => eth1g_reg_mosi,
    eth1g_reg_miso           => eth1g_reg_miso,
    eth1g_reg_interrupt      => eth1g_reg_interrupt,
    eth1g_ram_mosi           => eth1g_ram_mosi,
    eth1g_ram_miso           => eth1g_ram_miso,
        
    -- FPGA pins
    -- . General
    CLK                      => CLK,
    PPS                      => PPS,
    WDI                      => WDI,
    INTA                     => INTA,
    INTB                     => INTB,
    -- . Others
    VERSION                  => VERSION,
    ID                       => ID,
    TESTIO                   => TESTIO,
    -- . I2C Interface to Sensors
    sens_sc                  => sens_sc,
    sens_sd                  => sens_sd,        
    -- . 1GbE Control Interface
    ETH_clk                  => ETH_clk,
    ETH_SGIN                 => ETH_SGIN,
    ETH_SGOUT                => ETH_SGOUT
  );

  -----------------------------------------------------------------------------
  -- MM master
  -----------------------------------------------------------------------------
  u_mmm : ENTITY work.mmm_unb_ddr3_transpose
  GENERIC MAP (
    g_sim           => g_sim,
    g_sim_unb_nr    => g_sim_unb_nr,
    g_sim_node_nr   => g_sim_node_nr,
    g_frame_size_in => c_frame_size_in,
    g_nof_streams   => c_nof_streams,
    g_ddr3_seq      => c_ddr3_seq_conf
   )
  PORT MAP(  
    xo_clk                    => xo_clk,       
    xo_rst_n                  => xo_rst_n,     
    xo_rst                    => xo_rst,       
                              
    mm_rst                    => mm_rst,
    mm_clk                    => mm_clk,       
    mm_locked                 => mm_locked,    
                              
    -- PIOs                   
    pout_wdi                  => pout_wdi,
                              
    -- Manual WDI override    
    reg_wdi_mosi              => reg_wdi_mosi,
    reg_wdi_miso              => reg_wdi_miso,
                              
    -- system_info            
    reg_unb_system_info_mosi  => reg_unb_system_info_mosi,
    reg_unb_system_info_miso  => reg_unb_system_info_miso,
    rom_unb_system_info_mosi  => rom_unb_system_info_mosi,
    rom_unb_system_info_miso  => rom_unb_system_info_miso, 
                              
    -- UniBoard I2C sensors   
    reg_unb_sens_mosi         => reg_unb_sens_mosi,
    reg_unb_sens_miso         => reg_unb_sens_miso, 
                              
    -- PPSH                   
    reg_ppsh_mosi             => reg_ppsh_mosi,
    reg_ppsh_miso             => reg_ppsh_miso, 
                              
    -- eth1g                  
    eth1g_tse_clk             => eth1g_tse_clk,
    eth1g_mm_rst              => eth1g_mm_rst,
    eth1g_tse_mosi            => eth1g_tse_mosi,
    eth1g_tse_miso            => eth1g_tse_miso,
    eth1g_reg_mosi            => eth1g_reg_mosi,
    eth1g_reg_miso            => eth1g_reg_miso,
    eth1g_reg_interrupt       => eth1g_reg_interrupt,
    eth1g_ram_mosi            => eth1g_ram_mosi,
    eth1g_ram_miso            => eth1g_ram_miso,
                              
    -- Blockgenerator         
    reg_diag_bg_mosi          => reg_diag_bg_mosi,
    reg_diag_bg_miso          => reg_diag_bg_miso,
    ram_diag_bg_mosi          => ram_diag_bg_mosi,
    ram_diag_bg_miso          => ram_diag_bg_miso,
                              
    -- DDR3 transpose         
    ram_ss_ss_transp_mosi     => ram_ss_ss_transp_mosi,
    ram_ss_ss_transp_miso     => ram_ss_ss_transp_miso,
                             
    -- Databuffers            
    ram_diag_data_buf_im_mosi => ram_diag_data_buf_im_mosi,
    ram_diag_data_buf_im_miso => ram_diag_data_buf_im_miso,
    reg_diag_data_buf_im_mosi => reg_diag_data_buf_im_mosi,
    reg_diag_data_buf_im_miso => reg_diag_data_buf_im_miso,
    ram_diag_data_buf_re_mosi => ram_diag_data_buf_re_mosi,
    ram_diag_data_buf_re_miso => ram_diag_data_buf_re_miso,
    reg_diag_data_buf_re_mosi => reg_diag_data_buf_re_mosi,
    reg_diag_data_buf_re_miso => reg_diag_data_buf_re_miso,
    
    -- BSN monitor
    reg_bsn_monitor_mosi      => reg_bsn_monitor_mosi,
    reg_bsn_monitor_miso      => reg_bsn_monitor_miso,
    
    reg_io_ddr_mosi           => reg_io_ddr_mosi,   
    reg_io_ddr_miso           => reg_io_ddr_miso
    
  );

  -----------------------------------------------------------------------------
  -- Node function
  -----------------------------------------------------------------------------
  u_bg : ENTITY diag_lib.mms_diag_block_gen
  GENERIC MAP(
    g_nof_streams        => c_nof_streams,
    g_buf_dat_w          => c_bg_buf_dat_w,
    g_buf_addr_w         => c_bg_buf_adr_w,            
    g_file_index_arr     => c_bg_data_file_index_arr,
    g_file_name_prefix   => c_bg_data_file_prefix
  )
  PORT MAP(
    -- System
    mm_rst               => mm_rst,
    mm_clk               => mm_clk,
    dp_rst               => dp_rst,
    dp_clk               => dp_clk,
    en_sync              => dp_pps,
    -- MM interface      
    reg_bg_ctrl_mosi     => reg_diag_bg_mosi,
    reg_bg_ctrl_miso     => reg_diag_bg_miso,
    ram_bg_data_mosi     => ram_diag_bg_mosi,
    ram_bg_data_miso     => ram_diag_bg_miso,
    -- ST interface      
    out_siso_arr         => bg_siso_arr,
    out_sosi_arr         => bg_sosi_arr
  );
  
  u_bsn_monitor : ENTITY dp_lib.mms_dp_bsn_monitor
  GENERIC MAP (
    g_nof_streams        => 2, -- Check one input and one output stream
    g_cross_clock_domain => TRUE,
    g_bsn_w              => c_dp_stream_bsn_w,
    g_cnt_sop_w          => c_word_w,
    g_cnt_valid_w        => c_word_w,
    g_log_first_bsn      => TRUE
  )
  PORT MAP (
    -- Memory-mapped clock domain
    mm_rst      => mm_rst,
    mm_clk      => mm_clk,
    reg_mosi    => reg_bsn_monitor_mosi,
    reg_miso    => reg_bsn_monitor_miso,
    
    -- Streaming clock domain
    dp_rst      => dp_rst,
    dp_clk      => dp_clk,
    in_siso_arr => (OTHERS=>c_dp_siso_rdy),
    in_sosi_arr => bsn_sosi_arr
  );
  
  bsn_sosi_arr(0) <= bg_sosi_arr(0);
  bsn_sosi_arr(1) <= out_sosi_arr(0);
    

  u_areset_ddr_ref_rst : ENTITY common_lib.common_areset
  GENERIC MAP(
    g_rst_level => '1',
    g_delay_len => 40
  )
  PORT MAP(
    clk     => CLK,
    in_rst  => mm_rst,
    out_rst => ddr_ref_rst
  );    


  u_ddr3_T: ENTITY ddr3_lib.ddr3_transpose
  GENERIC MAP(
    g_sim                 => g_sim,                         
    g_nof_streams         => c_nof_streams,      
    g_in_dat_w            => c_bg_buf_dat_w/c_nof_complex, 
    g_frame_size_in       => c_frame_size_in,               
    g_frame_size_out      => c_frame_size_out, 
    g_nof_blk_per_sync    => c_nof_blk_per_sync,
    g_use_complex         => TRUE,   
    g_ena_pre_transp      => c_ena_pre_transpose,                    
    g_phy                 => c_phy,                     
    g_mts                 => c_mts,                     
    g_ddr3_seq            => c_ddr3_seq_conf
  )                          
  PORT MAP (        
    mm_rst                => mm_rst, 
    mm_clk                => mm_clk,

    dp_ref_clk            => CLK, 
    dp_ref_rst            => ddr_ref_rst,

    dp_rst                => dp_rst, 
    dp_clk                => dp_clk,

    dp_out_clk            => dp_clk, 
    dp_out_rst            => dp_rst,
    
    reg_io_ddr_mosi       => reg_io_ddr_mosi,   
    reg_io_ddr_miso       => reg_io_ddr_miso,   
                          
    snk_out_arr           => bg_siso_arr,
    snk_in_arr            => bg_sosi_arr,
    -- ST source          
    src_in_arr            => out_siso_arr,
    src_out_arr           => out_sosi_arr,
    
    ram_ss_ss_transp_mosi => ram_ss_ss_transp_mosi,
    ram_ss_ss_transp_miso => ram_ss_ss_transp_miso,
    
    ser_term_ctrl_out     => OPEN,
    par_term_ctrl_out     => OPEN,
                          
    ser_term_ctrl_in      => OPEN,
    par_term_ctrl_in      => OPEN,
                          
    phy_in                => MB_I_in,
    phy_io                => MB_I_io,     
    phy_ou                => MB_I_ou
  );

  ----------------------------------------------------------------------------
  -- Sink: data buffer real 
  ---------------------------------------------------------------------------- 
  u_data_buf_re : ENTITY diag_lib.mms_diag_data_buffer
  GENERIC MAP (    
    g_nof_streams     => c_nof_streams, 
    g_data_type       => c_db_data_type_re,   
    g_data_w          => c_db_data_w,      
    g_buf_nof_data    => c_db_buf_nof_data,
    g_buf_use_sync    => c_db_buf_use_sync
  )
  PORT MAP (
    -- System
    mm_rst            => mm_rst,
    mm_clk            => mm_clk,
    dp_rst            => dp_rst,
    dp_clk            => dp_clk,
     -- MM interface
    ram_data_buf_mosi => ram_diag_data_buf_re_mosi,
    ram_data_buf_miso => ram_diag_data_buf_re_miso,
    reg_data_buf_mosi => reg_diag_data_buf_re_mosi,
    reg_data_buf_miso => reg_diag_data_buf_re_miso,
    -- ST interface
    in_sync           => OPEN,
    in_sosi_arr       => out_sosi_arr
  );

  ----------------------------------------------------------------------------
  -- Sink: data buffer imag 
  ---------------------------------------------------------------------------- 
  u_data_buf_im : ENTITY diag_lib.mms_diag_data_buffer
  GENERIC MAP (    
    g_nof_streams     => c_nof_streams, 
    g_data_type       => c_db_data_type_im,   
    g_data_w          => c_db_data_w,      
    g_buf_nof_data    => c_db_buf_nof_data,
    g_buf_use_sync    => c_db_buf_use_sync
  )
  PORT MAP (
    -- System
    mm_rst            => mm_rst,
    mm_clk            => mm_clk,
    dp_rst            => dp_rst,
    dp_clk            => dp_clk,
    -- MM interface
    ram_data_buf_mosi => ram_diag_data_buf_im_mosi,
    ram_data_buf_miso => ram_diag_data_buf_im_miso,
    reg_data_buf_mosi => reg_diag_data_buf_im_mosi,
    reg_data_buf_miso => reg_diag_data_buf_im_miso,
    -- ST interface
    in_sync           => OPEN,
    in_sosi_arr       => out_sosi_arr
  );

END str;

