-------------------------------------------------------------------------------
--
-- Copyright (C) 2012
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- JIVE (Joint Institute for VLBI in Europe) <http://www.jive.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

-- Purpose: Verify that unb1_terminal_bg_mesh_db can simulate
-- Description: No verification and no need to run a main.c. Just check that it
--   loads OK and runs OK (i.e. that the DUT IO are toggling) for a few us.
-- Usage:
--   > as 3
--   > run 10 us

LIBRARY IEEE, common_lib, unb1_board_lib;
USE IEEE.std_logic_1164.ALL;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.tb_common_pkg.ALL;
USE unb1_board_lib.unb1_board_pkg.ALL;

ENTITY tb_unb1_terminal_bg_mesh_db IS
END tb_unb1_terminal_bg_mesh_db;

ARCHITECTURE tb OF tb_unb1_terminal_bg_mesh_db IS

  CONSTANT c_sim             : BOOLEAN := TRUE;
  CONSTANT c_version         : STD_LOGIC_VECTOR(1 DOWNTO 0) := "00";
  CONSTANT c_id_fn0          : STD_LOGIC_VECTOR(7 DOWNTO 0) := "00000100";
  CONSTANT c_id_bn3          : STD_LOGIC_VECTOR(7 DOWNTO 0) := "00000111";

  CONSTANT c_sim_unb_nr      : NATURAL := 0;
  CONSTANT c_sim_node_nr     : NATURAL := 0;
  
  CONSTANT c_cable_delay     : TIME := 12 ns;
  
  CONSTANT c_ext_clk_period  : TIME := 5 ns;  -- 200 MHz
  CONSTANT c_eth_clk_period  : TIME := 40 ns;  -- 25 MHz XO on UniBoard                                                                             
  CONSTANT c_tr_clk_period   : TIME := 6400 ps;  -- 156.25 MHz XO on UniBoard                                                                             
  
  -- DUT
  SIGNAL WDI                 : STD_LOGIC;
  SIGNAL ext_pps             : STD_LOGIC;
  SIGNAL ext_clk             : STD_LOGIC := '0';
  SIGNAL INTA                : STD_LOGIC;
  SIGNAL INTB                : STD_LOGIC;
  
  SIGNAL VERSION             : STD_LOGIC_VECTOR(c_unb1_board_aux.version_w-1 DOWNTO 0);
  SIGNAL ID                  : STD_LOGIC_VECTOR(c_unb1_board_aux.id_w-1 DOWNTO 0);
  SIGNAL TESTIO              : STD_LOGIC_VECTOR(c_unb1_board_aux.testio_w-1 DOWNTO 0);
  
  SIGNAL sens_scl            : STD_LOGIC;
  SIGNAL sens_sda            : STD_LOGIC;
  
  SIGNAL eth_clk             : STD_LOGIC := '0';
  SIGNAL eth_txp             : STD_LOGIC;
  SIGNAL eth_rxp             : STD_LOGIC;
  
  SIGNAL tr_clk              : STD_LOGIC := '0';
   
  SIGNAL FN_BN_0_TX          : STD_LOGIC_VECTOR(c_unb1_board_tr_mesh.bus_w-1 DOWNTO 0);
  SIGNAL FN_BN_0_RX          : STD_LOGIC_VECTOR(c_unb1_board_tr_mesh.bus_w-1 DOWNTO 0);
  SIGNAL FN_BN_1_TX          : STD_LOGIC_VECTOR(c_unb1_board_tr_mesh.bus_w-1 DOWNTO 0);
  SIGNAL FN_BN_1_RX          : STD_LOGIC_VECTOR(c_unb1_board_tr_mesh.bus_w-1 DOWNTO 0);
  SIGNAL FN_BN_2_TX          : STD_LOGIC_VECTOR(c_unb1_board_tr_mesh.bus_w-1 DOWNTO 0);
  SIGNAL FN_BN_2_RX          : STD_LOGIC_VECTOR(c_unb1_board_tr_mesh.bus_w-1 DOWNTO 0);
  SIGNAL FN_BN_3_TX          : STD_LOGIC_VECTOR(c_unb1_board_tr_mesh.bus_w-1 DOWNTO 0);
  SIGNAL FN_BN_3_RX          : STD_LOGIC_VECTOR(c_unb1_board_tr_mesh.bus_w-1 DOWNTO 0);

BEGIN

  -- Run 1 ms
  
  ext_clk <= NOT ext_clk AFTER c_ext_clk_period/2;  -- External clock (200 MHz) 
  eth_clk <= NOT eth_clk AFTER c_eth_clk_period/2;  -- 1GbE XO clock (25 MHz) 
  tr_clk  <= NOT tr_clk  AFTER c_tr_clk_period/2;    -- Transceiver clock (156.25 MHz) 
  
  ext_pps <= '0';   -- not used
  
  INTA <= 'H';  -- pull up
  INTB <= 'H';  -- pull up
  
  sens_scl <= 'H';  -- pull up
  sens_sda <= 'H';  -- pull up

  VERSION <= c_version;
  ID      <= c_id_fn0;
  
  ------------------------------------------------------------------------------
  -- UniBoard FPGA Node
  ------------------------------------------------------------------------------
  u_dut : ENTITY work.unb1_terminal_bg_mesh_db 
  GENERIC MAP (
    -- General
    g_sim         => c_sim,
    g_sim_level   => 1,
    g_sim_unb_nr  => c_sim_unb_nr,
    g_sim_node_nr => c_sim_node_nr
  )
  PORT MAP (
    -- GENERAL
    WDI           => WDI,
    CLK           => ext_clk, 
    PPS           => ext_pps,
    INTA          => INTA,
    INTB          => INTB,
    
    -- Others
    VERSION       => VERSION,
    ID            => ID,
    TESTIO        => TESTIO,

    -- I2C Interface to Sensors
    sens_sc       => sens_scl, 
    sens_sd       => sens_sda,
    
    -- 1GbE Control Interface
    ETH_clk       => eth_clk,   -- ETH reference clock also used for system reference clock
    ETH_SGIN      => eth_rxp,
    ETH_SGOUT     => eth_txp,
    
    -- Transceiver clocks
    SB_CLK        => tr_clk,  -- TR clock FN-BN(mesh)

    -- Mesh serial I/O
    FN_BN_0_TX    => FN_BN_0_TX,
    FN_BN_0_RX    => FN_BN_0_RX,
    FN_BN_1_TX    => FN_BN_1_TX,
    FN_BN_1_RX    => FN_BN_1_RX,
    FN_BN_2_TX    => FN_BN_2_TX,
    FN_BN_2_RX    => FN_BN_2_RX,
    FN_BN_3_TX    => FN_BN_3_TX,
    FN_BN_3_RX    => FN_BN_3_RX
  );
 
  ------------------------------------------------------------------------------
  -- Loopback model
  ------------------------------------------------------------------------------
  
  eth_rxp    <= TRANSPORT eth_txp AFTER c_cable_delay;
  
  FN_BN_0_RX <= TRANSPORT FN_BN_0_TX AFTER c_cable_delay;
  FN_BN_1_RX <= TRANSPORT FN_BN_1_TX AFTER c_cable_delay;
  FN_BN_2_RX <= TRANSPORT FN_BN_2_TX AFTER c_cable_delay;
  FN_BN_3_RX <= TRANSPORT FN_BN_3_TX AFTER c_cable_delay;
 
 
END tb;
