-------------------------------------------------------------------------------
--
-- Copyright (C) 2011
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- JIVE (Joint Institute for VLBI in Europe) <http://www.jive.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

-- Purpose: Compensate for the mesh reorder between FN and BN on UniBoard
-- Description: See unb1_board_mesh_reorder_bidir.vhd
--   This unb1_board_mesh_reorder_tx performs mesh reordering for user transmit,
--   so for the tx_usr_sosi and the rx_usr_siso. Note that these belong to
--   different streams. The tx_usr_sosi carries the user Tx data output stream
--   and the rx_usr_siso belongs to the user Rx data stream and carries the 
--   output flow control for the Rx data stream.
-- Remark:
-- . Indexing for *_2arr = (node id 0,1,2,3)(tr lane 3,2,1,0)

LIBRARY IEEE, common_lib, dp_lib;
USE IEEE.std_logic_1164.ALL;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;
USE work.unb1_board_pkg.ALL;


ENTITY unb1_board_mesh_reorder_tx IS
  GENERIC (
    g_node_type : t_e_unb1_board_node := e_any;
    g_reorder   : BOOLEAN := TRUE
  );
  PORT (
    chip_id          : IN  STD_LOGIC_VECTOR(c_unb1_board_aux.chip_id_w-1 DOWNTO 0);  -- [2:0]
    clk              : IN  STD_LOGIC;
    tx_usr_sosi_2arr : IN  t_unb1_board_mesh_sosi_2arr;  -- _2arr = (node id 0,1,2,3)(tr lane 3,2,1,0)
    rx_usr_siso_2arr : IN  t_unb1_board_mesh_siso_2arr := c_unb1_board_mesh_siso_2arr_rst;
    tx_phy_sosi_2arr : OUT t_unb1_board_mesh_sosi_2arr;
    rx_phy_siso_2arr : OUT t_unb1_board_mesh_siso_2arr
  );
END unb1_board_mesh_reorder_tx;


ARCHITECTURE rtl OF unb1_board_mesh_reorder_tx IS
  
  SIGNAL chip_id_reg : STD_LOGIC_VECTOR(chip_id'RANGE);
  SIGNAL chip_id_i   : STD_LOGIC_VECTOR(chip_id'RANGE);
  
BEGIN

  -- Register the chip_id from FPGA pins to ease timing closure.
  -- . Alternatively these registers may better be removed and pin input chip_id[] set as false path for timing closure
  u_chip_id : ENTITY common_lib.common_pipeline
  GENERIC MAP (
    g_representation => "UNSIGNED",
    g_pipeline       => c_meta_delay_len,
    g_reset_value    => 0,
    g_in_dat_w       => chip_id'LENGTH,
    g_out_dat_w      => chip_id'LENGTH
  )
  PORT MAP (
    rst     => '0',
    clk     => clk,
    in_dat  => chip_id,
    out_dat => chip_id_reg
  );
  
  -- force chip_id(2) to '0' or '1' to reduce the case options in p_comb if the design will run only on a FN or only on a BN
  chip_id_i <= func_unb1_board_chip_id(chip_id_reg, g_node_type);
  
  p_comb : PROCESS(chip_id_i, tx_usr_sosi_2arr, rx_usr_siso_2arr)
  BEGIN
    -- default connect to node 0,1,2,3 via port 0,1,2,3
    tx_phy_sosi_2arr <= tx_usr_sosi_2arr;
    rx_phy_siso_2arr <= rx_usr_siso_2arr;
    IF g_reorder=TRUE THEN
      -- make the reordering to compensate for the reordering of the UniBoard mesh
      CASE TO_UINT(chip_id_i) IS
        WHEN 0 => -- this is FN0, connect usr bus 0,1,2,3 to phy bus 0,3,2,1
          -- sosi
          tx_phy_sosi_2arr(3) <= tx_usr_sosi_2arr(1);
          tx_phy_sosi_2arr(2) <= tx_usr_sosi_2arr(2);
          tx_phy_sosi_2arr(1) <= tx_usr_sosi_2arr(3);
          -- siso
          rx_phy_siso_2arr(3) <= rx_usr_siso_2arr(1);
          rx_phy_siso_2arr(2) <= rx_usr_siso_2arr(2);
          rx_phy_siso_2arr(1) <= rx_usr_siso_2arr(3);
        WHEN 1 => -- this is FN1, connect usr bus 0,1,2,3 to phy bus 0,1,3,2
          -- sosi
          tx_phy_sosi_2arr(3) <= tx_usr_sosi_2arr(2);
          tx_phy_sosi_2arr(2) <= tx_usr_sosi_2arr(3);
          -- siso
          rx_phy_siso_2arr(3) <= rx_usr_siso_2arr(2);
          rx_phy_siso_2arr(2) <= rx_usr_siso_2arr(3);
        WHEN 2 => -- this is FN2, connect usr bus 0,1,2,3 to phy bus 0,1,2,3
        WHEN 3 => -- this is FN3, connect usr bus 0,1,2,3 to phy bus 1,0,2,3
          -- sosi
          tx_phy_sosi_2arr(1) <= tx_usr_sosi_2arr(0);
          tx_phy_sosi_2arr(0) <= tx_usr_sosi_2arr(1);
          -- siso
          rx_phy_siso_2arr(1) <= rx_usr_siso_2arr(0);
          rx_phy_siso_2arr(0) <= rx_usr_siso_2arr(1);
        WHEN 4 |
             5 => -- this is BN0
                  --      or BN1, connect usr bus 0,1,2,3 to phy bus 3,2,1,0
          -- sosi
          tx_phy_sosi_2arr(3) <= tx_usr_sosi_2arr(0);
          tx_phy_sosi_2arr(2) <= tx_usr_sosi_2arr(1);
          tx_phy_sosi_2arr(1) <= tx_usr_sosi_2arr(2);
          tx_phy_sosi_2arr(0) <= tx_usr_sosi_2arr(3);
          -- siso
          rx_phy_siso_2arr(3) <= rx_usr_siso_2arr(0);
          rx_phy_siso_2arr(2) <= rx_usr_siso_2arr(1);
          rx_phy_siso_2arr(1) <= rx_usr_siso_2arr(2);
          rx_phy_siso_2arr(0) <= rx_usr_siso_2arr(3);
        WHEN 6 => -- this is BN2, connect usr bus 0,1,2,3 to phy bus 2,3,1,0
          -- sosi
          tx_phy_sosi_2arr(2) <= tx_usr_sosi_2arr(0);
          tx_phy_sosi_2arr(3) <= tx_usr_sosi_2arr(1);
          tx_phy_sosi_2arr(1) <= tx_usr_sosi_2arr(2);
          tx_phy_sosi_2arr(0) <= tx_usr_sosi_2arr(3);
          -- siso
          rx_phy_siso_2arr(2) <= rx_usr_siso_2arr(0);
          rx_phy_siso_2arr(3) <= rx_usr_siso_2arr(1);
          rx_phy_siso_2arr(1) <= rx_usr_siso_2arr(2);
          rx_phy_siso_2arr(0) <= rx_usr_siso_2arr(3);
        WHEN 7 => -- this is BN3, connect usr bus 0,1,2,3 to phy bus 1,3,2,0
          -- sosi
          tx_phy_sosi_2arr(1) <= tx_usr_sosi_2arr(0);
          tx_phy_sosi_2arr(3) <= tx_usr_sosi_2arr(1);
          tx_phy_sosi_2arr(2) <= tx_usr_sosi_2arr(2);
          tx_phy_sosi_2arr(0) <= tx_usr_sosi_2arr(3);
          -- siso
          rx_phy_siso_2arr(1) <= rx_usr_siso_2arr(0);
          rx_phy_siso_2arr(3) <= rx_usr_siso_2arr(1);
          rx_phy_siso_2arr(2) <= rx_usr_siso_2arr(2);
          rx_phy_siso_2arr(0) <= rx_usr_siso_2arr(3);
        WHEN OTHERS => NULL;
      END CASE;
    END IF;
  END PROCESS;
  
END rtl;
