-------------------------------------------------------------------------------
--
-- Copyright (C) 2012
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- JIVE (Joint Institute for VLBI in Europe) <http://www.jive.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

-- Purpose: Model the UniBoard backplane of the Apertif beamformer
-- Description: 
--   The Apertif backplane connects 4 UniBoards using serial links. All BN0 are
--   connected to all other BN0 in a full mesh of transceiver lanes. Idem for
--   all BN1, for all BN2 and for all BN3. Each point to point bus between two
--   BN contains 4 transceiver lanes.
--
--   UniBoard bck_id=3 connects to UniBoards 2,1,0 via phy busses 2,1,0
--                   2                       3,1,0                1,0,2
--                   1                       3,2,0                0,1,2
--                   0                       3,2,1                0,2,1
-- Remark:
-- . Phy bus 3 is not used and left not connected on the backplane.

LIBRARY IEEE, common_lib, dp_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE dp_lib.dp_stream_pkg.ALL;
USE common_lib.common_pkg.ALL;
USE work.unb1_board_pkg.ALL;
USE work.tb_unb1_board_pkg.ALL;

ENTITY unb1_board_back_model_sl IS
  PORT (
    backplane_in_serial_4arr  : IN  t_unb1_board_back_sl_4arr;  -- _4arr = (UNB 3,2,1,0)(BN 3,2,1,0)(phy bus 2,1,0)(tr lane 3,2,1,0)
    backplane_out_serial_4arr : OUT t_unb1_board_back_sl_4arr
  );
END unb1_board_back_model_sl;


ARCHITECTURE beh OF unb1_board_back_model_sl IS

BEGIN

  gen_bn : FOR BN IN 0 TO c_unb1_board_nof_bn-1 GENERATE
                 
    ----------------------------------------------------------------------------
    -- Interconnect BN --> BN streaming with actual backplane routing
    ----------------------------------------------------------------------------
    
    backplane_out_serial_4arr(0)(BN)(0) <= backplane_in_serial_4arr(3)(BN)(0);
    backplane_out_serial_4arr(0)(BN)(1) <= backplane_in_serial_4arr(1)(BN)(2);
    backplane_out_serial_4arr(0)(BN)(2) <= backplane_in_serial_4arr(2)(BN)(2);
    
    backplane_out_serial_4arr(1)(BN)(0) <= backplane_in_serial_4arr(3)(BN)(1);
    backplane_out_serial_4arr(1)(BN)(1) <= backplane_in_serial_4arr(2)(BN)(0);
    backplane_out_serial_4arr(1)(BN)(2) <= backplane_in_serial_4arr(0)(BN)(1);
    
    backplane_out_serial_4arr(2)(BN)(0) <= backplane_in_serial_4arr(1)(BN)(1);
    backplane_out_serial_4arr(2)(BN)(1) <= backplane_in_serial_4arr(3)(BN)(2);
    backplane_out_serial_4arr(2)(BN)(2) <= backplane_in_serial_4arr(0)(BN)(2);
    
    backplane_out_serial_4arr(3)(BN)(0) <= backplane_in_serial_4arr(0)(BN)(0);
    backplane_out_serial_4arr(3)(BN)(1) <= backplane_in_serial_4arr(1)(BN)(0);
    backplane_out_serial_4arr(3)(BN)(2) <= backplane_in_serial_4arr(2)(BN)(1); 
    --                        ^   ^  ^                              ^      ^
    --                        |   |  |                              |      |
    --                        |   |  BN_BI RX phy bus               |      BN_BI TX phy bus
    --                        |   |                                 Transmitting UniBoard
    --                        |   Same scheme applies to all back nodes
    --                        Receiving UniBoard     

  END GENERATE;
  
END beh;
