-------------------------------------------------------------------------------
--
-- Copyright (C) 2012
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- JIVE (Joint Institute for VLBI in Europe) <http://www.jive.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

-- Purpose: Test bench for unb2_test.
-- Description:
--   The DUT can be targeted at unb 0, node 3 with the same Python scripts 
--   that are used on hardware. 
-- Usage:
--   On command line do:
--     > run_modelsim & (to start Modeslim)
--
--   In Modelsim do:
--     > lp unb2_test
--     > mk clean all (only first time to clean all libraries)
--     > mk all (to compile all libraries that are needed for unb2_test)
--     . load tb_unb1_test simulation by double clicking the tb_unb2_test icon
--     > as 10 (to view signals in Wave Window)
--     > run 100 us (or run -all)
--
--   On command line do:
--     > python $UPE/peripherals/util_system_info.py --gn 3 -n 0 -v 5 --sim
--     > python $UPE/peripherals/util_unb_sens.py --gn 3 -n 0 -v 5 --sim
--     > python $UPE/peripherals/util_ppsh.py --gn 3 -n 1 -v 5 --sim
--

LIBRARY IEEE, common_lib, unb2_board_lib, i2c_lib, technology_lib, tech_pll_lib;
USE IEEE.std_logic_1164.ALL;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE unb2_board_lib.unb2_board_pkg.ALL;
USE common_lib.tb_common_pkg.ALL;
USE technology_lib.technology_pkg.ALL;
USE tech_pll_lib.tech_pll_component_pkg.ALL;

ENTITY tb_unb2_test IS
    GENERIC (
      g_design_name : STRING  := "unb2_test";
      g_sim_unb_nr  : NATURAL := 0; -- UniBoard 0
      g_sim_node_nr : NATURAL := 3  -- Node 3
    );
END tb_unb2_test;

ARCHITECTURE tb OF tb_unb2_test IS

  CONSTANT c_sim             : BOOLEAN := TRUE;

  CONSTANT c_unb_nr          : NATURAL := 0; -- UniBoard 0
  CONSTANT c_node_nr         : NATURAL := 3; -- Node 3
  CONSTANT c_id              : STD_LOGIC_VECTOR(7 DOWNTO 0) := TO_UVEC(c_unb_nr, c_unb2_board_nof_uniboard_w) & TO_UVEC(c_node_nr, c_unb2_board_nof_chip_w);

  CONSTANT c_version         : STD_LOGIC_VECTOR(1 DOWNTO 0) := "00";
  CONSTANT c_fw_version      : t_unb2_board_fw_version := (1, 0);

  CONSTANT c_cable_delay     : TIME := 12 ns;
  CONSTANT c_eth_clk_period  : TIME := 8 ns;  -- 125 MHz XO on UniBoard 
  CONSTANT c_clk_period      : TIME := 5 ns; 
  CONSTANT c_sa_clk_period   : TIME := tech_pll_clk_644_period; -- 644 MHz
  CONSTANT c_sb_clk_period   : TIME := tech_pll_clk_644_period; -- 644 MHz
  CONSTANT c_bck_ref_clk_period : TIME := tech_pll_clk_644_period; -- 644 MHz
  CONSTANT c_pps_period      : NATURAL := 1000; 

  -- DUT
  SIGNAL clk                 : STD_LOGIC := '0';
  SIGNAL pps                 : STD_LOGIC := '0';
  SIGNAL pps_rst             : STD_LOGIC := '0';

  SIGNAL WDI                 : STD_LOGIC;
  SIGNAL INTA                : STD_LOGIC;
  SIGNAL INTB                : STD_LOGIC;

  SIGNAL eth_clk             : STD_LOGIC := '0';
  SIGNAL eth_txp             : STD_LOGIC_VECTOR(1 DOWNTO 0);
  SIGNAL eth_rxp             : STD_LOGIC_VECTOR(1 DOWNTO 0);
  
  SIGNAL VERSION             : STD_LOGIC_VECTOR(c_unb2_board_aux.version_w-1 DOWNTO 0) := c_version; 
  SIGNAL ID                  : STD_LOGIC_VECTOR(c_unb2_board_aux.id_w-1 DOWNTO 0)      := c_id;
  SIGNAL TESTIO              : STD_LOGIC_VECTOR(c_unb2_board_aux.testio_w-1 DOWNTO 0);

  SIGNAL sens_scl            : STD_LOGIC;
  SIGNAL sens_sda            : STD_LOGIC;

  SIGNAL qsfp_led            : STD_LOGIC_VECTOR(c_unb2_board_tr_qsfp_nof_leds-1 DOWNTO 0);

  -- 10GbE
  SIGNAL sa_clk              : STD_LOGIC := '1';
  SIGNAL sb_clk              : STD_LOGIC := '1';
  SIGNAL bck_ref_clk         : STD_LOGIC := '1';

  -- Serial I/O
  SIGNAL si_lpbk_0           : STD_LOGIC_VECTOR(c_unb2_board_tr_qsfp.bus_w-1 DOWNTO 0);
  SIGNAL si_lpbk_1           : STD_LOGIC_VECTOR(c_unb2_board_tr_qsfp.bus_w-1 DOWNTO 0);
  SIGNAL si_lpbk_2           : STD_LOGIC_VECTOR(c_unb2_board_tr_qsfp.bus_w-1 DOWNTO 0);
  SIGNAL si_lpbk_3           : STD_LOGIC_VECTOR(c_unb2_board_tr_qsfp.bus_w-1 DOWNTO 0);
  SIGNAL si_lpbk_4           : STD_LOGIC_VECTOR(c_unb2_board_tr_qsfp.bus_w-1 DOWNTO 0);
  SIGNAL si_lpbk_5           : STD_LOGIC_VECTOR(c_unb2_board_tr_qsfp.bus_w-1 DOWNTO 0);

  SIGNAL si_lpbk_6           : STD_LOGIC_VECTOR(c_unb2_board_tr_ring.bus_w-1 DOWNTO 0);
  SIGNAL si_lpbk_7           : STD_LOGIC_VECTOR(c_unb2_board_tr_ring.bus_w-1 DOWNTO 0);

  SIGNAL si_lpbk_8           : STD_LOGIC_VECTOR(c_unb2_board_tr_back.bus_w-1 DOWNTO 0);


  -- Model I2C sensor slaves as on the UniBoard
  CONSTANT c_fpga_temp_address   : STD_LOGIC_VECTOR(6 DOWNTO 0) := "0011000";  -- MAX1618 address LOW LOW
  CONSTANT c_fpga_temp           : INTEGER := 60;
  CONSTANT c_eth_temp_address    : STD_LOGIC_VECTOR(6 DOWNTO 0) := "0101001";  -- MAX1618 address MID LOW
  CONSTANT c_eth_temp            : INTEGER := 40;
  CONSTANT c_hot_swap_address    : STD_LOGIC_VECTOR(6 DOWNTO 0) := "1000100";  -- LTC4260 address L L L
  CONSTANT c_hot_swap_R_sense    : REAL := 0.01;                               -- = 10 mOhm on UniBoard
  
  CONSTANT c_uniboard_current    : REAL := 5.0;   -- = assume 5.0 A on UniBoard
  CONSTANT c_uniboard_supply     : REAL := 48.0;  -- = assume 48.0 V on UniBoard
  CONSTANT c_uniboard_adin       : REAL := -1.0;  -- = NC on UniBoard

BEGIN

  ----------------------------------------------------------------------------
  -- System setup
  ----------------------------------------------------------------------------
  clk     <= NOT clk AFTER c_clk_period/2;        -- External clock (200 MHz)
  eth_clk <= NOT eth_clk AFTER c_eth_clk_period/2;  -- Ethernet ref clock (125 MHz)
  sa_clk  <= NOT sa_clk  AFTER c_sa_clk_period/2;   -- sa clock (644 MHz)
  sb_clk  <= NOT sb_clk  AFTER c_sb_clk_period/2;   -- sb clock (644 MHz)
  bck_ref_clk  <= NOT bck_ref_clk  AFTER c_bck_ref_clk_period/2;  -- bck_ref clock (644 MHz)

  
  INTA <= 'H';  -- pull up
  INTB <= 'H';  -- pull up

  sens_scl <= 'H';  -- pull up
  sens_sda <= 'H';  -- pull up

  ------------------------------------------------------------------------------
  -- External PPS
  ------------------------------------------------------------------------------  
  proc_common_gen_pulse(1, c_pps_period, '1', pps_rst, clk, pps);

  ------------------------------------------------------------------------------
  -- 1GbE Loopback model
  ------------------------------------------------------------------------------  
  eth_rxp(0) <= TRANSPORT eth_txp(0) AFTER c_cable_delay;

  eth_rxp(1) <= '0';

  eth_txp(1) <= '0';
  
  ------------------------------------------------------------------------------
  -- DUT
  ------------------------------------------------------------------------------
  u_unb2_test : ENTITY work.unb2_test
    GENERIC MAP (
      g_sim         => c_sim,
      g_sim_unb_nr  => c_unb_nr,
      g_sim_node_nr => c_node_nr,
      g_design_name => g_design_name
    )
    PORT MAP (
      -- GENERAL
      CLK         => clk,
      PPS         => pps,
      WDI         => WDI,
      INTA        => INTA,
      INTB        => INTB,

      SENS_SC     => sens_scl,
      SENS_SD     => sens_sda,

      -- Others
      VERSION     => VERSION,
      ID          => ID,
      TESTIO      => TESTIO,

      -- 1GbE Control Interface
      ETH_CLK     => eth_clk,
      ETH_SGIN    => eth_rxp,
      ETH_SGOUT   => eth_txp,

      -- Transceiver clocks
      SA_CLK      => sa_clk,
      SB_CLK      => sb_clk,
      BCK_REF_CLK => bck_ref_clk,

      PMBUS_ALERT => '0',

      -- Serial I/O
      QSFP_0_TX  => si_lpbk_0,
      QSFP_0_RX  => si_lpbk_0,
--      QSFP_1_TX  => si_lpbk_1,
--      QSFP_1_RX  => si_lpbk_1,
--      QSFP_2_TX  => si_lpbk_2,
--      QSFP_2_RX  => si_lpbk_2,
--      QSFP_3_TX  => si_lpbk_3,
--      QSFP_3_RX  => si_lpbk_3,
--      QSFP_4_TX  => si_lpbk_4,
--      QSFP_4_RX  => si_lpbk_4,
--      QSFP_5_TX  => si_lpbk_5,
--      QSFP_5_RX  => si_lpbk_5,
--
--      RING_0_TX  => si_lpbk_6,
--      RING_0_RX  => si_lpbk_6,
--      RING_1_TX  => si_lpbk_7,
--      RING_1_RX  => si_lpbk_7,
--
--      BCK_TX     => si_lpbk_8,
--      BCK_RX     => si_lpbk_8,

      QSFP_LED   => qsfp_led
    );

  ------------------------------------------------------------------------------
  -- UniBoard sensors
  ------------------------------------------------------------------------------
  -- I2C slaves that are available for each FPGA
  u_fpga_temp : ENTITY i2c_lib.dev_max1618
  GENERIC MAP (
    g_address => c_fpga_temp_address
  )
  PORT MAP (
    scl  => sens_scl,
    sda  => sens_sda,
    temp => c_fpga_temp
  );

  -- I2C slaves that are available only via FPGA back node 3
  u_eth_temp : ENTITY i2c_lib.dev_max1618
  GENERIC MAP (
    g_address => c_eth_temp_address
  )
  PORT MAP (
    scl  => sens_scl,
    sda  => sens_sda,
    temp => c_eth_temp
  );
  
  u_power : ENTITY i2c_lib.dev_ltc4260
  GENERIC MAP (
    g_address => c_hot_swap_address,
    g_R_sense => c_hot_swap_R_sense
  )
  PORT MAP (
    scl               => sens_scl,
    sda               => sens_sda,
    ana_current_sense => c_uniboard_current,
    ana_volt_source   => c_uniboard_supply,
    ana_volt_adin     => c_uniboard_adin
  );

END tb;
