--------------------------------------------------------------------------------
--
-- Copyright (C) 2010
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- JIVE (Joint Institute for VLBI in Europe) <http://www.jive.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------
 
LIBRARY IEEE, common_lib;
USE IEEE.std_logic_1164.ALL;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;


-- Purpose: Generate a stream of frames with test sequence data.
-- Description:
--   Each frame has g_frame_len words of out_dat. The test data can be PRSG or
--   COUNTER dependent on diag_sel, or constant data when diag_dc='1'.
--   The frame generator is enabled by diag_en. The actual frame rate depends
--   on the g_sop_period and on out_ready.
--   The out_ready acts as a data request. During a frame the out_dat is valid
--   for every active out_ready, to support streaming flow control.
-- Remark:
-- . This component can be easily mapped to standard MM and ST interfaces.


ENTITY diag_frm_generator IS
  GENERIC (
    g_sel         : STD_LOGIC := '1';  -- '0' = PRSG, '1' = COUNTER
    g_frame_len   : NATURAL := 2;      -- >= 2, test frame length in nof dat words
    g_sof_period  : NATURAL := 1;      -- >= 1, nof cycles between sop that start a frame, typically >> g_frame_len
                                       --       to generate frames with idle in between
    g_frame_cnt_w : NATURAL := 32;
    g_dat_w       : NATURAL := 16;     -- >= 1, test data width
    g_symbol_w    : NATURAL := 16;     -- >= 1, and nof_symbols_per_dat = g_dat_w/g_symbol_w, must be an integer
    g_empty       : NATURAL := 0       -- >= 0 and < nof symbols per dat, 
  );
  PORT (
    rst              : IN  STD_LOGIC;
    clk              : IN  STD_LOGIC;
    clken            : IN  STD_LOGIC := '1';
    
    -- Static control input (connect via MM or leave open to use default)
    diag_en          : IN  STD_LOGIC;
    diag_sel         : IN  STD_LOGIC := g_sel;  -- '0' = PRSG sequence data, '1' = COUNTER sequence data
    diag_dc          : IN  STD_LOGIC := '0';    -- '0' = output diag_sel sequence data, '1' = output constant diag_init data
    diag_frame_len   : IN  STD_LOGIC_VECTOR(ceil_log2(g_frame_len)-1 DOWNTO 0) := TO_UVEC(g_frame_len, ceil_log2(g_frame_len));
    diag_frame_empty : IN  STD_LOGIC_VECTOR(ceil_log2(g_dat_w/g_symbol_w)-1 DOWNTO 0) := TO_UVEC(g_empty, ceil_log2(g_dat_w/g_symbol_w));
    diag_sof_period  : IN  STD_LOGIC_VECTOR(ceil_log2(g_sof_period)-1 DOWNTO 0) := TO_UVEC(g_sof_period, ceil_log2(g_sof_period));
    diag_frame_cnt   : OUT STD_LOGIC_VECTOR(g_frame_cnt_w-1 DOWNTO 0);
    
    -- ST output
    out_ready       : IN  STD_LOGIC := '1';                      -- '1' = request ST test data output, '0' = halt ST test data output
    out_dat         : OUT STD_LOGIC_VECTOR(g_dat_w-1 DOWNTO 0);  -- test sequence data
    out_val         : OUT STD_LOGIC;                             -- '1' when out_data is valid
    out_sop         : OUT STD_LOGIC;                             -- '1' for first valid frame data
    out_eop         : OUT STD_LOGIC;                             -- '1' for last  valid frame data
    out_empty       : OUT STD_LOGIC_VECTOR(ceil_log2(g_dat_w/g_symbol_w)-1 DOWNTO 0)  -- nof empty symbols in last out_dat word marked by out_eop
  );
END diag_frm_generator;


ARCHITECTURE str OF diag_frm_generator IS

  CONSTANT c_init           : NATURAL := 0;  -- first data word of first frame that is generated after diag_en = '1'
  CONSTANT c_frame_len_w    : NATURAL := ceil_log2(g_frame_len);
  
  SIGNAL diag_en_revt       : STD_LOGIC;
  SIGNAL diag_init          : STD_LOGIC_VECTOR(g_dat_w-1 DOWNTO 0) := TO_UVEC(c_init, g_dat_w);  -- init data for out_dat when diag_sop = '1'
  SIGNAL nxt_diag_init      : STD_LOGIC_VECTOR(diag_init'RANGE);
  SIGNAL diag_sop           : STD_LOGIC;
  SIGNAL diag_ready         : STD_LOGIC;
  
  SIGNAL frame_cnt          : STD_LOGIC_VECTOR(diag_frame_cnt'RANGE);
  SIGNAL nxt_diag_frame_cnt : STD_LOGIC_VECTOR(diag_frame_cnt'RANGE);
  
  SIGNAL i_out_sop          : STD_LOGIC;
  SIGNAL i_out_eop          : STD_LOGIC;
  
BEGIN

  out_sop   <= i_out_sop;
  out_eop   <= i_out_eop;
  out_empty <= diag_frame_empty;
  
  -- Signal begin of diag_en
  u_diag_en_revt : ENTITY common_lib.common_evt
  GENERIC MAP (
    g_evt_type => "RISING",
    g_out_reg  => FALSE
  )
  PORT MAP (
    rst     => rst,
    clk     => clk,
    clken   => clken,
    in_sig  => diag_en,
    out_evt => diag_en_revt
  );
  
  p_clk : PROCESS (rst, clk)
  BEGIN
    IF rst='1' THEN
      diag_init      <= TO_UVEC(c_init, g_dat_w);
      diag_frame_cnt <= (OTHERS=>'0');
    ELSIF rising_edge(clk) THEN
      IF clken='1' THEN
        diag_init      <= nxt_diag_init;
        diag_frame_cnt <= nxt_diag_frame_cnt;
      END IF;
    END IF;
  END PROCESS;
  
  nxt_diag_init <= TO_UVEC(c_init, g_dat_w) WHEN diag_en='0' ELSE INCR_UVEC(diag_init, 1) WHEN i_out_sop='1';
  
  u_pulse_sop : ENTITY common_lib.common_pulser
  GENERIC MAP (
    g_pulse_period => g_sof_period
  )
  PORT MAP (
    rst            => rst,
    clk            => clk,
    clken          => clken,
    pulse_period   => diag_sof_period,
    pulse_en       => diag_en,
    pulse_clr      => diag_en_revt,
    pulse_out      => diag_sop
  );
  
  -- Hold last frame count also when the generator is disabled, clear when it is restarted
  nxt_diag_frame_cnt <= (OTHERS=>'0') WHEN diag_en_revt='1' ELSE frame_cnt;
    
  u_frm_cnt : ENTITY common_lib.common_counter
  GENERIC MAP (
    g_width     => g_frame_cnt_w
  )
  PORT MAP (
    rst     => rst,
    clk     => clk,
    clken   => clken,
    cnt_clr => diag_en_revt,
    cnt_en  => i_out_eop,
    count   => frame_cnt
  );  
    
  u_tx_frm : ENTITY work.diag_tx_frm
  GENERIC MAP (
    g_sel       => g_sel,
    g_init      => c_init,
    g_frame_len => g_frame_len,
    g_dat_w     => g_dat_w
  )
  PORT MAP (
    rst            => rst,
    clk            => clk,
    clken          => clken,
    
    -- Static control input (connect via MM or leave open to use default)
    diag_sel       => diag_sel,
    diag_dc        => diag_dc,
    diag_frame_len => diag_frame_len,
    
    -- Dynamic control input (connect via MM or via ST input or leave open to use defaults)
    diag_ready     => diag_ready,
    diag_init      => diag_init,
    diag_sop       => diag_sop,
    
    -- ST output
    out_ready      => out_ready,
    out_dat        => out_dat,
    out_val        => out_val,
    out_sop        => i_out_sop,
    out_eop        => i_out_eop
  );
    
END str;
