--------------------------------------------------------------------------------
--
-- Copyright (C) 2012
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------

LIBRARY IEEE,common_lib;
USE IEEE.std_logic_1164.ALL;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE work.dp_stream_pkg.ALL;

-- Purpose: 
-- . Deinterleave the input into g_nof_out outputs, outputting g_block_size words before switching to next output.
-- Description: 
-- . DP wrapper for common_deinterleave
-- Remark:
-- . Usage is straightforward, but when g_use_ctrl is TRUE, do make sure the input blocks can be evenly
--   distributed over the chosen number of outputs at the chosen deinterleaver block size (g_block_size).

ENTITY dp_deinterleave IS
  GENERIC (
    g_dat_w          : NATURAL;
    g_nof_out        : NATURAL;
    g_block_size     : NATURAL;         -- Deinterleaver block size; output g_block_size words before witching to next output
    g_use_ctrl       : BOOLEAN := TRUE; -- Requires: [input block size (sop:eop)] / [g_nof_out]/ [g_block_size] = integer number!
    g_use_complex    : BOOLEAN;
    g_align_out      : BOOLEAN := FALSE 
  );
  PORT (
    rst         : IN  STD_LOGIC;
    clk         : IN  STD_LOGIC;

    snk_in      : IN  t_dp_sosi;
    src_out_arr : OUT t_dp_sosi_arr(g_nof_out-1 DOWNTO 0)
  );
END dp_deinterleave;


ARCHITECTURE wrap OF dp_deinterleave IS

  CONSTANT c_complex_w  : NATURAL := g_dat_w/2;

  -- Arrays flattened into SLVs
  SIGNAL out_dat        : STD_LOGIC_VECTOR(g_nof_out*g_dat_w-1 DOWNTO 0);
  SIGNAL out_val        : STD_LOGIC_VECTOR(g_nof_out-1 DOWNTO 0);

  SIGNAL piped_snk_in   : t_dp_sosi;

  -- Either SOSI data or SOSI Im&Re is forwarded to the deinterleaver
  SIGNAL in_dat         : STD_LOGIC_VECTOR(g_dat_w-1 DOWNTO 0);
  SIGNAL piped_in_dat   : STD_LOGIC_VECTOR(g_dat_w-1 DOWNTO 0);

BEGIN

  gen_sosi_dat_in: IF g_use_complex = FALSE GENERATE
    -----------------------------------------------------------------------------
    -- Forward SOSI data field
    -----------------------------------------------------------------------------
    in_dat <= snk_in.data(g_dat_w-1 DOWNTO 0);
  END GENERATE;

  gen_sosi_compl_in: IF g_use_complex = TRUE GENERATE
    -----------------------------------------------------------------------------
    -- Forward concatenated im & re fields
    -----------------------------------------------------------------------------
    in_dat <= snk_in.im(c_complex_w-1 DOWNTO 0) & snk_in.re(c_complex_w-1 DOWNTO 0);
  END GENERATE;

  u_deinterleave : ENTITY common_lib.common_deinterleave
  GENERIC MAP (
    g_nof_out    => g_nof_out,
    g_block_size => g_block_size,
    g_dat_w      => g_dat_w,
    g_align_out  => g_align_out
  )
  PORT MAP (
    rst        => rst,
    clk        => clk,
    
    in_dat     => piped_in_dat,
    in_val     => piped_snk_in.valid, 
    
    out_dat    => out_dat,
    out_val    => out_val
  );

  -----------------------------------------------------------------------------
  -- Map output SLV to sosi_arr
  -----------------------------------------------------------------------------
  gen_wires_out : FOR i IN 0 TO g_nof_out-1 GENERATE
    gen_sosi_dat_out: IF g_use_complex = FALSE GENERATE
      -----------------------------------------------------------------------------
      -- Forward SOSI data field
      -----------------------------------------------------------------------------
      src_out_arr(i).data(g_dat_w-1 DOWNTO 0) <= out_dat( i*g_dat_w+g_dat_w -1 DOWNTO i*g_dat_w);
    END GENERATE;
    
    gen_sosi_compl_out: IF g_use_complex = TRUE GENERATE
      -----------------------------------------------------------------------------
      -- Forward concatenated im & re fields
      -----------------------------------------------------------------------------
      src_out_arr(i).im <= RESIZE_DP_DSP_DATA(out_dat( i*g_dat_w + g_dat_w     -1 DOWNTO i*g_dat_w + c_complex_w));
      src_out_arr(i).re <= RESIZE_DP_DSP_DATA(out_dat( i*g_dat_w + c_complex_w -1 DOWNTO i*g_dat_w));
    END GENERATE;

    src_out_arr(i).valid <= out_val(i);
  END GENERATE;

  -----------------------------------------------------------------------------
  -- The input data and valid are delayed for g_block_size cycles before they
  -- enter common_deinterleave, so the snk_in.eop will be aligned with 
  -- src_out(0).eop.
  -----------------------------------------------------------------------------
  u_pipe_dat : ENTITY common_lib.common_pipeline
  GENERIC MAP (
    g_pipeline  => g_block_size,
    g_in_dat_w  => g_dat_w,
    g_out_dat_w => g_dat_w
  )
  PORT MAP (
    rst     => rst,
    clk     => clk,
    in_dat  => in_dat,
    out_dat => piped_in_dat 
  );

  u_pipe_val : ENTITY common_lib.common_pipeline
  GENERIC MAP (
    g_pipeline  => g_block_size,
    g_in_dat_w  => 1,
    g_out_dat_w => 1
  )
  PORT MAP (
    rst     => rst,
    clk     => clk,
    in_dat  => slv(snk_in.valid),
    sl(out_dat) => piped_snk_in.valid 
  );        
  
  -----------------------------------------------------------------------------
  -- Forward and align the input Sync, BSN, SOP and EOP with the outputs
  -----------------------------------------------------------------------------
  gen_ctrl : if g_use_ctrl=TRUE GENERATE
    gen_out_pipes: FOR i IN 0 TO g_nof_out-1 GENERATE
      -----------------------------------------------------------------------------
      -- Before forwarding the remaining (ctrl) signals to the outputs, they also 
      -- need the base delay of g_block_size*, plus additional incremental delays 
      -- (i*g_block_size) to align them with the data and valid (that already have
      -- those delays because of the sequential deinterleaving) of the 
      -- corresponding output (i).
      --
      -- *this excludes eop as that was the reason for delaying the other inputs
      --  by g_block_size cycles in the first place.
      -----------------------------------------------------------------------------
      u_pipe_sync : ENTITY common_lib.common_pipeline
      GENERIC MAP (
        g_pipeline  => g_block_size + i*g_block_size,
        g_in_dat_w  => 1,
        g_out_dat_w => 1
      )
      PORT MAP (
        rst     => rst,
        clk     => clk,
        in_dat  => slv(snk_in.sync),
        sl(out_dat) => src_out_arr(i).sync
      );

      u_pipe_bsn : ENTITY common_lib.common_pipeline
      GENERIC MAP (
        g_pipeline  => g_block_size + i*g_block_size,
        g_in_dat_w  => c_dp_stream_bsn_w,
        g_out_dat_w => c_dp_stream_bsn_w
      )
      PORT MAP (
        rst     => rst,
        clk     => clk,
        in_dat  => snk_in.bsn,
        out_dat => src_out_arr(i).bsn
      );

      u_pipe_sop : ENTITY common_lib.common_pipeline
      GENERIC MAP (
        g_pipeline  => g_block_size + i*g_block_size,
        g_in_dat_w  => 1,
        g_out_dat_w => 1
      )
      PORT MAP (
        rst     => rst,
        clk     => clk,
        in_dat  => slv(snk_in.sop),
        sl(out_dat) => src_out_arr(i).sop
      );

      u_pipe_eop : ENTITY common_lib.common_pipeline
      GENERIC MAP (
        g_pipeline  => i*g_block_size,
        g_in_dat_w  => 1,
        g_out_dat_w => 1
      )
      PORT MAP (
        rst     => rst,
        clk     => clk,
        in_dat  => slv(snk_in.eop),
        sl(out_dat) => src_out_arr(i).eop
      );

      u_pipe_err : ENTITY common_lib.common_pipeline
      GENERIC MAP (
        g_pipeline  => g_block_size + i*g_block_size,
        g_in_dat_w  => c_dp_stream_error_w,
        g_out_dat_w => c_dp_stream_error_w
      )
      PORT MAP (
        rst     => rst,
        clk     => clk,
        in_dat  => snk_in.err,
        out_dat => src_out_arr(i).err
      );

      u_pipe_empty : ENTITY common_lib.common_pipeline
      GENERIC MAP (
        g_pipeline  => g_block_size + i*g_block_size,
        g_in_dat_w  => c_dp_stream_empty_w,
        g_out_dat_w => c_dp_stream_empty_w
      )
      PORT MAP (
        rst     => rst,
        clk     => clk,
        in_dat  => snk_in.empty,
        out_dat => src_out_arr(i).empty
      );

      u_pipe_channel : ENTITY common_lib.common_pipeline
      GENERIC MAP (
        g_pipeline  => g_block_size + i*g_block_size,
        g_in_dat_w  => c_dp_stream_channel_w,
        g_out_dat_w => c_dp_stream_channel_w
      )
      PORT MAP (
        rst     => rst,
        clk     => clk,
        in_dat  => snk_in.channel,
        out_dat => src_out_arr(i).channel
      );

    END GENERATE;
  END GENERATE;

END wrap;
