--------------------------------------------------------------------------------
--
-- Copyright (C) 2014
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------

LIBRARY IEEE, common_lib, dp_lib;
USE IEEE.std_logic_1164.ALL;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;

-- Purpose:
-- Description:

ENTITY corr_unfolder IS
  GENERIC (
    g_nof_inputs  : NATURAL;     -- Number of inputs
    g_nof_unfolds : NATURAL := 0 -- Number of times to unfold
  );
  PORT (
    rst            : IN  STD_LOGIC;
    clk            : IN  STD_LOGIC;

    snk_in_arr     : IN  t_dp_sosi_arr(g_nof_inputs-1 DOWNTO 0);

    src_out_arr    : OUT t_dp_sosi_arr(g_nof_inputs*pow2(g_nof_unfolds)-1 DOWNTO 0)
  );
END corr_unfolder;

ARCHITECTURE str OF corr_unfolder IS

  COMPONENT corr_unfolder IS
    GENERIC (
      g_nof_inputs  : NATURAL;
      g_nof_unfolds : NATURAL := 0 
    ); 
    PORT (
      rst            : IN  STD_LOGIC;
      clk            : IN  STD_LOGIC;
  
      snk_in_arr     : IN  t_dp_sosi_arr(g_nof_inputs-1 DOWNTO 0);
  
      src_out_arr    : OUT t_dp_sosi_arr(g_nof_inputs*pow2(g_nof_unfolds)-1 DOWNTO 0)
    );
  END COMPONENT;

  CONSTANT c_nof_demuxes : NATURAL := g_nof_inputs; 

  SIGNAL output_sel_arr           : STD_LOGIC_VECTOR(c_nof_demuxes-1 DOWNTO 0);
  SIGNAL nxt_output_sel_arr       : STD_LOGIC_VECTOR(c_nof_demuxes-1 DOWNTO 0);

  SIGNAL demux_src_out_2arr_2     : t_dp_sosi_2arr_2(c_nof_demuxes-1 DOWNTO 0);
  SIGNAL nxt_demux_src_out_2arr_2 : t_dp_sosi_2arr_2(c_nof_demuxes-1 DOWNTO 0);
  SIGNAL demux_src_out_arr        : t_dp_sosi_arr(2*c_nof_demuxes-1 DOWNTO 0);

BEGIN

  gen_arch: IF g_nof_unfolds/=0 GENERATE
    -----------------------------------------------------------------------------
    -- Simple 2-output demux logic
    -----------------------------------------------------------------------------
    gen_demux_comb: FOR i IN 0 TO c_nof_demuxes-1 GENERATE
      nxt_output_sel_arr(i) <= NOT output_sel_arr(i) WHEN snk_in_arr(i).valid='1' ELSE output_sel_arr(i);

      nxt_demux_src_out_2arr_2(i)(0) <= snk_in_arr(i) WHEN output_sel_arr(i)='0' AND snk_in_arr(i).valid='1' ELSE c_dp_sosi_rst;
      nxt_demux_src_out_2arr_2(i)(1) <= snk_in_arr(i) WHEN output_sel_arr(i)='1' AND snk_in_arr(i).valid='1' ELSE c_dp_sosi_rst;
    END GENERATE;

    -----------------------------------------------------------------------------
    -- Wire the 2D demux output array to 1D array to match entity I/O type
    -----------------------------------------------------------------------------
    gen_demux_inputs_0: FOR i IN 0 TO c_nof_demuxes-1 GENERATE
       demux_src_out_arr(2*i)   <= demux_src_out_2arr_2(i)(0);
       demux_src_out_arr(2*i+1) <= demux_src_out_2arr_2(i)(1);
    END GENERATE;
  
    -----------------------------------------------------------------------------
    -- If g_nof_unfolds=1, this is the last stage. Otherwise, add a stage.
    -----------------------------------------------------------------------------
    gen_corr_folder: IF g_nof_unfolds>1 GENERATE
      u_corr_folder : corr_unfolder
      GENERIC MAP (
        g_nof_inputs  => c_nof_demuxes*2, -- Next stage has all our demux outputs as inputs
        g_nof_unfolds => g_nof_unfolds-1
      )
      PORT MAP (
        rst         => rst,
        clk         => clk,
                    
        snk_in_arr  => demux_src_out_arr,                   
        src_out_arr => src_out_arr
      );
    END GENERATE;
  
    gen_src_out_arr: IF g_nof_unfolds=1 GENERATE
      src_out_arr <= demux_src_out_arr;
    END GENERATE;
  
    -----------------------------------------------------------------------------
    -- Registers
    -----------------------------------------------------------------------------
    p_clk: PROCESS(clk, rst)
    BEGIN
      IF rst='1' THEN
        demux_src_out_2arr_2 <= (OTHERS=>(OTHERS=>c_dp_sosi_rst));
        output_sel_arr       <= (OTHERS=>'0');
      ELSIF rising_edge(clk) THEN
        demux_src_out_2arr_2 <= nxt_demux_src_out_2arr_2;
        output_sel_arr       <= nxt_output_sel_arr;
      END IF;
    END PROCESS;

  END GENERATE;

  -----------------------------------------------------------------------------
  -- Wire output to input if g_nof_unfolds=0
  -----------------------------------------------------------------------------
  gen_wire_out_to_in: IF g_nof_unfolds=0 GENERATE
    src_out_arr <= snk_in_arr;
  END GENERATE;

END str;

