--------------------------------------------------------------------------------
--
-- Copyright (C) 2014
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------

LIBRARY IEEE, common_lib, dp_lib, diag_lib, mm_lib;
USE IEEE.std_logic_1164.ALL;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;
USE common_lib.tb_common_pkg.ALL;
USE diag_lib.diag_pkg.ALL;
USE mm_lib.mm_file_pkg.ALL;
USE mm_lib.mm_file_unb_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;

-- Folding rules:
-- . g_nof_input_folds=g_nof_pre_mult_folds : no gaps are needed/desired in the input data.
-- . g_nof_input_folds<g_nof_pre_mult_folds : gaps are required in the input data
-- . g_nof_input_folds>g_nof_pre_mult_folds : is not allowed
--   . Produces unnecessary/undesired gaps in the multiplier output
ENTITY tb_correlator IS
  GENERIC (
    g_nof_inputs         : NATURAL := 8;
    g_nof_channels       : NATURAL := 8;
    g_nof_pre_mult_folds : NATURAL := 1;-- Make sure c_nof_visibilities can be folded this many times
    g_nof_input_folds    : NATURAL := 1 -- =< g_nof_pre_mult_folds. Use an even g_nof_inputs.
  );
END tb_correlator;

ARCHITECTURE tb OF tb_correlator IS

  CONSTANT c_complex_data_w     : NATURAL := 8;

  CONSTANT c_nof_input_streams  : NATURAL := g_nof_inputs / pow2(g_nof_input_folds);
  CONSTANT c_conjugate          : BOOLEAN := TRUE;
  CONSTANT c_integration_period : NATURAL := 0;
  CONSTANT c_nof_visibilities   : NATURAL := (g_nof_inputs*(g_nof_inputs+1))/2;

  CONSTANT c_input_gaps_needed  : BOOLEAN := g_nof_input_folds<g_nof_pre_mult_folds;
  CONSTANT c_block_period       : NATURAL := sel_a_b(c_input_gaps_needed, pow2(g_nof_pre_mult_folds-g_nof_input_folds), 1);

  -- Block generator
  CONSTANT c_bg_block_size      : NATURAL := g_nof_channels*pow2(g_nof_input_folds);
  CONSTANT c_bg_gapsize         : NATURAL := c_bg_block_size*(c_block_period-1);

  -- Indicate the integration period with the sync. In the correlator, the
  -- integration period is forced to a minimum of c_nof_visibilities to
  -- allow folding the output onto one stream.
  -- . The sync pulse is only there for the human eye (wave window) -
  --   it is not used by the correlator.
  CONSTANT c_bg_blocks_per_sync : NATURAL := largest(c_integration_period, c_nof_visibilities);

  CONSTANT c_bg_ctrl            : t_diag_block_gen := ('1',                      -- enable             
                                                       '0',                      -- enable_sync        
                                                      TO_UVEC(     c_bg_block_size, c_diag_bg_samples_per_packet_w),
                                                      TO_UVEC(c_bg_blocks_per_sync, c_diag_bg_blocks_per_sync_w),
                                                      TO_UVEC(        c_bg_gapsize, c_diag_bg_gapsize_w),
                                                      TO_UVEC(                   0, c_diag_bg_mem_low_adrs_w),
                                                      TO_UVEC(   c_bg_block_size-1, c_diag_bg_mem_high_adrs_w),
                                                      TO_UVEC(                   0, c_diag_bg_bsn_init_w));

  CONSTANT c_dp_clk_period      : TIME := 10 ns;
  CONSTANT c_mm_clk_period      : TIME := 10 ps;

  SIGNAL tb_end                 : STD_LOGIC := '0';
  SIGNAL dp_clk                 : STD_LOGIC := '1';
  SIGNAL dp_rst                 : STD_LOGIC;
  SIGNAL mm_clk                 : STD_LOGIC := '1';
  SIGNAL mm_rst                 : STD_LOGIC;

  SIGNAL block_gen_src_out_arr  : t_dp_sosi_arr(c_nof_input_streams-1 DOWNTO 0);

  SIGNAL dp_fifo_sc_src_out_arr : t_dp_sosi_arr(c_nof_input_streams-1 DOWNTO 0);
  SIGNAL dp_fifo_sc_src_in_arr  : t_dp_siso_arr(c_nof_input_streams-1 DOWNTO 0);

  SIGNAL correlator_snk_in_arr  : t_dp_sosi_arr(c_nof_input_streams-1 DOWNTO 0);
  SIGNAL correlator_src_out_arr : t_dp_sosi_arr(1-1 DOWNTO 0);

  SIGNAL ram_diag_data_buf_mosi : t_mem_mosi;
  SIGNAL ram_diag_data_buf_miso : t_mem_miso;
  SIGNAL reg_diag_data_buf_mosi : t_mem_mosi;
  SIGNAL reg_diag_data_buf_miso : t_mem_miso; 

  ----------------------------------------------------------------------------
  -- mm_file component
  ----------------------------------------------------------------------------
  COMPONENT mm_file
  GENERIC(
    g_file_prefix       : STRING;
    g_mm_clk_period     : TIME := c_mm_clk_period;
    g_update_on_change  : BOOLEAN := FALSE;
    g_mm_rd_latency     : NATURAL := 1
  );
  PORT (
    mm_rst        : IN  STD_LOGIC;
    mm_clk        : IN  STD_LOGIC;
    mm_master_out : OUT t_mem_mosi;
    mm_master_in  : IN  t_mem_miso 
  );
  END COMPONENT;

BEGIN

  -----------------------------------------------------------------------------
  -- Clocks and reset
  -----------------------------------------------------------------------------
  dp_clk <= NOT dp_clk OR tb_end AFTER c_dp_clk_period/2;
  dp_rst <= '1', '0' AFTER c_dp_clk_period*7;

  mm_clk <= NOT mm_clk OR tb_end AFTER c_mm_clk_period/2;
  mm_rst <= '1', '0' AFTER c_mm_clk_period*7;

  -----------------------------------------------------------------------------
  -- Block generators
  -- . Each stream in block_gen_src_out_arr contains complex subband samples
  --   from one 'antenna'.
  -- . These complex subband samples are generated and converted to a HEX RAM
  --   initialization file using Python, see tb/python/gen_subband_hex_files.py
  -----------------------------------------------------------------------------
  u_mms_diag_block_gen : ENTITY diag_lib.mms_diag_block_gen
  GENERIC MAP (
    g_nof_streams        => c_nof_input_streams,
    g_buf_dat_w          => 2*c_complex_data_w,
    g_buf_addr_w         => ceil_log2(TO_UINT(c_bg_ctrl.samples_per_packet)),
    g_file_name_prefix   => "../../../libraries/dsp/correlator/src/hex/complex_subbands_" & NATURAL'IMAGE(c_complex_data_w) & "b_" & "fold_" & NATURAL'IMAGE(g_nof_input_folds),
    g_diag_block_gen_rst => c_bg_ctrl
  )
  PORT MAP (
    mm_rst           => mm_rst,
    mm_clk           => mm_clk,

    dp_rst           => dp_rst,
    dp_clk           => dp_clk,

    out_sosi_arr     => block_gen_src_out_arr
  );

  -----------------------------------------------------------------------------
  -- Introduce gaps in the input stream
  -- . mms_diag_block_gen does not support gaps within blocks.
  -- . We'll use FIFO buffers and dp_gap to read out the FIFOs to introduce
  --   gaps.
  -----------------------------------------------------------------------------
  gen_dp_fifo_sc : FOR i IN 0 TO c_nof_input_streams-1 GENERATE
    u_dp_fifo_sc : ENTITY dp_lib.dp_fifo_sc
    GENERIC MAP (
      g_data_w         => 2*c_complex_data_w,
      g_use_ctrl       => FALSE,
      g_use_complex    => TRUE,
      g_fifo_size      => c_bg_block_size,
      g_fifo_af_margin => 0
    )
    PORT MAP (
      rst         => dp_rst,
      clk         => dp_clk,

      wr_ful      => OPEN,
      usedw       => OPEN,
      rd_emp      => OPEN,

      snk_out     => OPEN,
      snk_in      => block_gen_src_out_arr(i),
      src_in      => dp_fifo_sc_src_in_arr(i),
      src_out     => dp_fifo_sc_src_out_arr(i)
    );
  END GENERATE;

  gen_dp_src_out_timer : FOR i IN 0 TO c_nof_input_streams-1 GENERATE
    u_dp_src_out_timer : ENTITY dp_lib.dp_src_out_timer
    GENERIC MAP (
      g_block_period => c_block_period
    )
    PORT MAP (
      rst                  => dp_rst,
      clk                  => dp_clk,

      snk_in               => dp_fifo_sc_src_out_arr(i),
      snk_out              => dp_fifo_sc_src_in_arr(i)
    );
  END GENERATE;

  correlator_snk_in_arr <= dp_fifo_sc_src_out_arr;

  -----------------------------------------------------------------------------
  -- Device under test: correlator
  -----------------------------------------------------------------------------
  u_correlator : ENTITY work.correlator
  GENERIC MAP (
    g_sim                 => TRUE,
    g_nof_input_streams   => c_nof_input_streams,
    g_input_unfold_factor => g_nof_input_folds,
    g_nof_pre_mult_folds  => g_nof_pre_mult_folds,
    g_data_w              => c_complex_data_w,
    g_conjugate           => c_conjugate,
    g_nof_channels        => g_nof_channels,
    g_integration_period  => c_integration_period
  )
  PORT MAP (
    clk         => dp_clk,
    rst         => dp_rst,

    snk_in_arr  => correlator_snk_in_arr,
    src_out_arr => correlator_src_out_arr
  );

  -----------------------------------------------------------------------------
  -- Data buffers to be read out by Python
  -----------------------------------------------------------------------------
  u_diag_data_buffer : ENTITY diag_lib.mms_diag_data_buffer
  GENERIC MAP (    
    g_nof_streams  => 1,
    g_data_w       => 64,
    g_data_type    => e_complex,
    g_buf_nof_data => c_nof_visibilities,
    g_buf_use_sync => TRUE 
  )
  PORT MAP (
    mm_rst            => mm_rst,
    mm_clk            => mm_clk,
    dp_rst            => dp_rst,
    dp_clk            => dp_clk,

    ram_data_buf_mosi => ram_diag_data_buf_mosi,
    ram_data_buf_miso => ram_diag_data_buf_miso,
    reg_data_buf_mosi => reg_diag_data_buf_mosi,
    reg_data_buf_miso => reg_diag_data_buf_miso,

    in_sync           => correlator_src_out_arr(0).sop,
    in_sosi_arr       => correlator_src_out_arr
  );

  -----------------------------------------------------------------------------
  -- Stream recorder to record the correlator output stream to a file
  -- . The data buffer can only take snapshots.
  -----------------------------------------------------------------------------
  u_dp_stream_rec_play : ENTITY dp_lib.dp_stream_rec_play
  GENERIC MAP (
    g_sim            => TRUE,
    g_pass_through   => FALSE,
    g_rec_not_play   => TRUE,
    g_rec_play_file  => "$RADIOHDL/libraries/dsp/correlator/tb/rec/correlator_src_out_arr0.rec",
    g_record_invalid => FALSE
  )
  PORT MAP (
    dp_clk  => dp_clk,
    snk_in  => correlator_src_out_arr(0),
    snk_out => OPEN,
    src_out => OPEN,
    src_in  => c_dp_siso_rdy
  );

  -----------------------------------------------------------------------------
  -- MM file I/O with Python
  -----------------------------------------------------------------------------
  u_mm_file_ram_diag_data_buffer : mm_file GENERIC MAP(mmf_unb_file_prefix(0, 0, "FN") & "RAM_DIAG_DATA_BUFFER")
                                              PORT MAP(mm_rst, mm_clk, ram_diag_data_buf_mosi, ram_diag_data_buf_miso);

  u_mm_file_reg_diag_data_buffer : mm_file GENERIC MAP(mmf_unb_file_prefix(0, 0, "FN") & "REG_DIAG_DATA_BUFFER")
                                              PORT MAP(mm_rst, mm_clk, reg_diag_data_buf_mosi, reg_diag_data_buf_miso);

  ----------------------------------------------------------------------------
  -- Procedure that polls a sim control file that can be used to e.g. get
  -- the simulation time in ns
  ----------------------------------------------------------------------------
  mmf_poll_sim_ctrl_file(c_mmf_unb_file_path & "sim.ctrl", c_mmf_unb_file_path & "sim.stat");

END tb;
