-------------------------------------------------------------------------------
--
-- Copyright (C) 2012
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

LIBRARY IEEE, common_lib, dp_lib, tech_ddr_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;
USE tech_ddr_lib.tech_ddr_pkg.ALL;
USE work.ddr3_pkg.ALL;

ENTITY mms_ddr3_capture IS
  GENERIC (
    g_sim                     : BOOLEAN := FALSE; 
    g_ddr                     : t_c_ddr3_phy := c_ddr3_phy_4g;
    g_mts                     : NATURAL := 800;
    g_wr_data_w               : NATURAL := c_ddr3_ctlr_data_w; 
    g_wr_use_ctrl             : BOOLEAN := FALSE;
    g_wr_fifo_depth           : NATURAL := 128;
    g_rd_fifo_depth           : NATURAL := 256;  
    g_rd_data_w               : NATURAL := c_ddr3_ctlr_data_w;  
    g_flush_wr_fifo           : BOOLEAN := FALSE;  
    g_flush_sop               : BOOLEAN := FALSE;  
    g_flush_sop_channel       : BOOLEAN := FALSE;  
    g_flush_sop_start_channel : NATURAL := 0; 
    g_flush_nof_channels      : NATURAL := 0;
    g_use_bsn_scheduler       : BOOLEAN := FALSE
  );
  PORT (
    mm_rst          : IN    STD_LOGIC;
    mm_clk          : IN    STD_LOGIC;

    ctlr_ref_clk    : IN    STD_LOGIC;
    ctlr_rst        : IN    STD_LOGIC; -- asynchronous reset input to controller

    ctlr_gen_clk    : OUT   STD_LOGIC; -- Controller generated clock
    ctlr_gen_rst    : OUT   STD_LOGIC;    

    wr_clk          : IN    STD_LOGIC;
    wr_rst          : IN    STD_LOGIC;

    wr_sosi         : IN    t_dp_sosi;
    wr_siso         : OUT   t_dp_siso;
    
    flush_ena       : IN    STD_LOGIC := '0';
 
    -- MM registers
    ddr3_mosi       : IN    t_mem_mosi := c_mem_mosi_rst;
    ddr3_miso       : OUT   t_mem_miso;
  
    dpmm_ctrl_mosi  : IN    t_mem_mosi := c_mem_mosi_rst;
    dpmm_ctrl_miso  : OUT   t_mem_miso := c_mem_miso_rst;

    dpmm_data_mosi  : IN    t_mem_mosi := c_mem_mosi_rst;
    dpmm_data_miso  : OUT   t_mem_miso := c_mem_miso_rst;
    
    -- SO-DIMM Memory Bank
    ddr3_in         : IN    t_tech_ddr3_phy_in;
    ddr3_io         : INOUT t_tech_ddr3_phy_io;
    ddr3_ou         : OUT   t_tech_ddr3_phy_ou
  );
END mms_ddr3_capture;


ARCHITECTURE str OF mms_ddr3_capture IS
 
  SIGNAL rd_sosi    : t_dp_sosi;
  SIGNAL rd_siso    : t_dp_siso;
  
  SIGNAL rd_clk     : STD_LOGIC;
  SIGNAL rd_rst     : STD_LOGIC;

  SIGNAL rd_fifo_usedw : STD_LOGIC_VECTOR(ceil_log2(g_rd_fifo_depth * (c_ddr3_ctlr_data_w/g_rd_data_w) )-1 DOWNTO 0);

BEGIN

  u_mms_ddr3: ENTITY work.mms_ddr3
  GENERIC MAP(
    g_sim                      => g_sim,
    g_ddr                      => g_ddr,
    g_mts                      => g_mts,
    g_wr_data_w                => g_wr_data_w,
    g_wr_use_ctrl              => g_wr_use_ctrl,      
    g_wr_fifo_depth            => g_wr_fifo_depth,
    g_rd_fifo_depth            => g_rd_fifo_depth,     
    g_rd_data_w                => c_word_w, 
    g_flush_wr_fifo            => g_flush_wr_fifo,
    g_flush_sop                => g_flush_sop,
    g_flush_sop_channel        => g_flush_sop_channel,
    g_flush_sop_start_channel  => g_flush_sop_start_channel,
    g_flush_nof_channels       => g_flush_nof_channels
  )                          
  PORT MAP (        
    mm_rst             => mm_rst, 
    mm_clk             => mm_clk,
            
    ctlr_ref_clk       => ctlr_ref_clk,
    ctlr_rst           => ctlr_rst,

    ctlr_gen_clk       => ctlr_gen_clk,
    ctlr_gen_rst       => ctlr_gen_rst,      

    wr_clk             => wr_clk,
    wr_rst             => wr_rst,

    wr_sosi            => wr_sosi, 
    wr_siso            => wr_siso,
    
    flush_ena          => flush_ena,
  
    rd_sosi            => rd_sosi,
    rd_siso            => rd_siso,

    rd_clk             => mm_clk,
    rd_rst             => mm_rst,

    rd_fifo_usedw      => rd_fifo_usedw, -- relative to FIFO wr side

    ctrl_mosi          => ddr3_mosi,
    ctrl_miso          => ddr3_miso,

    ddr3_in            => ddr3_in,
    ddr3_io            => ddr3_io,     
    ddr3_ou            => ddr3_ou
  );

  u_mms_dp_fifo_to_mm: ENTITY dp_lib.mms_dp_fifo_to_mm
  GENERIC MAP(
    g_rd_fifo_depth => g_rd_fifo_depth * (c_ddr3_ctlr_data_w/g_rd_data_w)
  )                          
  PORT MAP (        
    mm_rst             => mm_rst, 
    mm_clk             => mm_clk,        
 
    rd_sosi            => rd_sosi,
    rd_siso            => rd_siso,

    ctrl_mosi          => dpmm_ctrl_mosi,
    ctrl_miso          => dpmm_ctrl_miso,

    data_mosi          => dpmm_data_mosi,
    data_miso          => dpmm_data_miso,

    rd_usedw           => rd_fifo_usedw

  );


END str;

