
-------------------------------------------------------------------------------
--
-- Copyright (C) 2012
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

LIBRARY IEEE, common_lib, dp_lib, tech_ddr_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;
USE tech_ddr_lib.tech_ddr_pkg.ALL;
USE work.ddr3_pkg.ALL;

ENTITY seq_ddr3 IS
  GENERIC (
    g_ddr      : t_c_ddr3_phy := c_ddr3_phy_4g;    
    g_phy      : NATURAL      := 1; 
    g_mts      : NATURAL      := 800;
    g_data_w   : NATURAL      := c_ddr3_ctlr_data_w; 
    g_ddr3_seq : t_ddr3_seq   := c_ddr3_seq
  );
  PORT (
    ctlr_ref_clk      : IN    STD_LOGIC;
    ctlr_rst          : IN    STD_LOGIC; -- asynchronous reset input to controller
                      
    ctlr_gen_clk      : OUT   STD_LOGIC; -- Controller generated clock
    ctlr_gen_rst      : OUT   STD_LOGIC;    
                      
    wr_clk            : IN    STD_LOGIC;
    wr_rst            : IN    STD_LOGIC;
                      
    wr_sosi           : IN    t_dp_sosi;
    wr_siso           : OUT   t_dp_siso;  
    
    flush_ena         : IN    STD_LOGIC;
                      
    rd_sosi           : OUT   t_dp_sosi;
    rd_siso           : IN    t_dp_siso;
                      
    rd_clk            : IN    STD_LOGIC;
    rd_rst            : IN    STD_LOGIC;
                      
    rd_data_mosi      : IN    t_mem_mosi := c_mem_mosi_rst;
    rd_data_miso      : OUT   t_mem_miso := c_mem_miso_rst;   
    
    ser_term_ctrl_out : OUT   STD_LOGIC_VECTOR(13 DOWNTO 0);
    par_term_ctrl_out : OUT   STD_LOGIC_VECTOR(13 DOWNTO 0);
                      
    ser_term_ctrl_in  : IN    STD_LOGIC_VECTOR(13 DOWNTO 0) := (OTHERS => '0');
    par_term_ctrl_in  : IN    STD_LOGIC_VECTOR(13 DOWNTO 0) := (OTHERS => '0');

    -- SO-DIMM Memory Bank
    ddr3_in          : IN    t_tech_ddr3_phy_in;
    ddr3_io          : INOUT t_tech_ddr3_phy_io;
    ddr3_ou          : OUT   t_tech_ddr3_phy_ou
  );
END seq_ddr3;


ARCHITECTURE str OF seq_ddr3 IS
  
  CONSTANT c_min_fifo_size    : POSITIVE := 256;
  CONSTANT c_blocksize        : POSITIVE := g_ddr3_seq.wr_nof_chunks * g_ddr3_seq.wr_chunksize;  
  CONSTANT c_wr_fifo_depth    : NATURAL  := sel_a_b(c_blocksize > c_min_fifo_size, c_blocksize, c_min_fifo_size);--c_blocksize * 2;
  CONSTANT c_rd_fifo_depth    : NATURAL  := sel_a_b(c_blocksize > c_min_fifo_size, c_blocksize, c_min_fifo_size);--c_blocksize * 2;
                              
  SIGNAL i_ctlr_gen_rst       : STD_LOGIC;
  SIGNAL i_ctlr_gen_clk       : STD_LOGIC;
  
  -- ctrl & status DDR3 driver
  SIGNAL dvr_start_addr       : t_ddr3_addr;
  SIGNAL dvr_end_addr         : t_ddr3_addr; 
                              
  SIGNAL dvr_en               : STD_LOGIC;
  SIGNAL dvr_wr_not_rd        : STD_LOGIC;
  SIGNAL dvr_done             : STD_LOGIC;        

  -- DDR3 controller status
  SIGNAL ctlr_init_done       : STD_LOGIC;    
  SIGNAL ctlr_rdy             : STD_LOGIC; 
  SIGNAL init_done_data_start : STD_LOGIC;  
  
BEGIN

  ctlr_gen_clk    <= i_ctlr_gen_clk;
  ctlr_gen_rst    <= i_ctlr_gen_rst;

  u_ddr3: ENTITY work.ddr3
  GENERIC MAP(
    g_ddr                     => g_ddr, 
    g_phy                     => g_phy,
    g_mts                     => g_mts,
    g_wr_data_w               => g_data_w,
    g_wr_use_ctrl             => TRUE,
    g_wr_fifo_depth           => c_wr_fifo_depth,
    g_rd_fifo_depth           => c_rd_fifo_depth,
    g_rd_data_w               => g_data_w,
    g_flush_wr_fifo           => TRUE,
    g_flush_sop               => TRUE,
    g_flush_sop_sync          => TRUE,
    g_flush_sop_channel       => FALSE,
    g_flush_sop_start_channel => 0,    
    g_flush_nof_channels      => 0     
  )                          
  PORT MAP (                      
    ctlr_ref_clk       => ctlr_ref_clk,
    ctlr_rst           => ctlr_rst,

    phy_in             => ddr3_in,
    phy_io             => ddr3_io,    
    phy_ou             => ddr3_ou,  

    ctlr_gen_clk       => i_ctlr_gen_clk,
    ctlr_gen_rst       => i_ctlr_gen_rst,      

    ctlr_init_done     => ctlr_init_done,

    ctlr_rdy           => ctlr_rdy,
    dvr_start_addr     => dvr_start_addr,
    dvr_end_addr       => dvr_end_addr,

    dvr_done           => dvr_done,
    dvr_wr_not_rd      => dvr_wr_not_rd,
    dvr_en             => dvr_en,
    
    wr_clk             => wr_clk,
    wr_rst             => wr_rst,

    wr_sosi            => wr_sosi, 
    wr_siso            => wr_siso,
    
    flush_ena          => flush_ena,
  
    rd_sosi            => rd_sosi,
    rd_siso            => rd_siso,

    rd_clk             => rd_clk,
    rd_rst             => rd_rst,    

    ser_term_ctrl_out  => ser_term_ctrl_out,
    par_term_ctrl_out  => par_term_ctrl_out,

    ser_term_ctrl_in   => ser_term_ctrl_in, 
    par_term_ctrl_in   => par_term_ctrl_in, 

    rd_fifo_usedw      => OPEN
  );
  
  init_done_data_start <= ctlr_init_done AND wr_sosi.sync; 

  u_ddr3_sequencer: ENTITY work.ddr3_seq
  GENERIC MAP(
    g_ddr      => g_ddr, 
    g_ddr3_seq => g_ddr3_seq
  )                      
  PORT MAP (       
    dp_rst     => wr_rst,
    dp_clk     => wr_clk,
    
    en_evt     => dvr_en, 
    wr_not_rd  => dvr_wr_not_rd,

    start_addr => dvr_start_addr,
    end_addr   => dvr_end_addr,

    done       => dvr_done,
    init_done  => init_done_data_start,
    ctlr_rdy   => ctlr_rdy
  );
    
END str;

