--------------------------------------------------------------------------------
--
-- Copyright (C) 2014
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- JIVE (Joint Institute for VLBI in Europe) <http://www.jive.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------

-- Purpose: Technology independent component for DDR memory access.
-- Description:
--   The component also supports different types of DDR, so DDR3 and DDR4.
-- Remark:

LIBRARY IEEE, common_lib, technology_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE technology_lib.technology_pkg.ALL;
USE technology_lib.technology_select_pkg.ALL;
USE work.tech_ddr_pkg.ALL;

ENTITY tech_ddr IS
  GENERIC (
    g_sim_model            : BOOLEAN := FALSE; -- TRUE: use fast behavioural model, requires no external memory (uses memory array).
    g_technology           : NATURAL := c_tech_select_default;
    g_tech_ddr             : t_c_tech_ddr
  );
  PORT (
    -- PLL reference clock
    ref_clk           : IN    STD_LOGIC;
    ref_rst           : IN    STD_LOGIC;

    -- Controller user interface
    ctlr_gen_clk      : OUT   STD_LOGIC;
    ctlr_gen_rst      : OUT   STD_LOGIC;
    ctlr_gen_clk_2x   : OUT   STD_LOGIC;
    ctlr_gen_rst_2x   : OUT   STD_LOGIC;

    ctlr_mosi         : IN    t_mem_ctlr_mosi;
    ctlr_miso         : OUT   t_mem_ctlr_miso;

    term_ctrl_out     : OUT   t_tech_ddr3_phy_terminationcontrol;
    term_ctrl_in      : IN    t_tech_ddr3_phy_terminationcontrol := c_tech_ddr3_phy_terminationcontrol_rst;

    -- DDR3 PHY interface
    phy3_in            : IN    t_tech_ddr3_phy_in := c_tech_ddr3_phy_in_x;
    phy3_io            : INOUT t_tech_ddr3_phy_io;
    phy3_ou            : OUT   t_tech_ddr3_phy_ou;
    
    -- DDR4 PHY interface
    phy4_in            : IN    t_tech_ddr4_phy_in := c_tech_ddr4_phy_in_x;
    phy4_io            : INOUT t_tech_ddr4_phy_io;
    phy4_ou            : OUT   t_tech_ddr4_phy_ou
  );
END tech_ddr;


ARCHITECTURE str OF tech_ddr IS

  CONSTANT c_nof_addr : NATURAL := 8192;
  CONSTANT c_dat_w    : NATURAL := 256;

  TYPE t_mem_arr IS ARRAY(0 TO c_nof_addr-1) OF STD_LOGIC_VECTOR(c_dat_w-1 DOWNTO 0);

  SIGNAL sim_clk : STD_LOGIC;
  SIGNAL sim_rst : STD_LOGIC;
  SIGNAL sim_ctlr_mosi : t_mem_ctlr_mosi;

  SIGNAL address : NATURAL;

BEGIN

  -----------------------------------------------------------------------------
  -- Technology IP cores
  ----------------------------------------------------------------------------- 
  gen_ip: IF g_sim_model = FALSE GENERATE
    gen_ip_stratixiv : IF g_technology=c_tech_stratixiv GENERATE
      u0 : ENTITY work.tech_ddr_stratixiv
      GENERIC MAP (g_tech_ddr)
      PORT MAP (ref_clk, ref_rst,
                ctlr_gen_clk, ctlr_gen_rst, ctlr_gen_clk_2x, ctlr_gen_rst_2x,
                ctlr_mosi, ctlr_miso, term_ctrl_out, term_ctrl_in,
                phy3_in, phy3_io, phy3_ou);
    END GENERATE;
    
    gen_ip_arria10 : IF g_technology=c_tech_arria10 GENERATE
      u0 : ENTITY work.tech_ddr_arria10
      GENERIC MAP (g_tech_ddr)
      PORT MAP (ref_clk, ref_rst,
                ctlr_gen_clk, ctlr_gen_rst,
                ctlr_mosi, ctlr_miso,
                phy4_in, phy4_io, phy4_ou);
    END GENERATE;
  END GENERATE;
 
  -----------------------------------------------------------------------------
  -- Functional simulation model
  -- . Note: is it assumed that the user only performs burst reads/writes!
  -----------------------------------------------------------------------------
  gen_sim_model : IF g_sim_model=TRUE GENERATE
    -- Prevent delta delay issues by using a re-assigned clk both internally 
    -- (sim_clk) and externally (ctrl_gen_clk)
    ctlr_gen_clk <= ref_clk;
    ctlr_gen_rst <= ref_rst;
    sim_clk      <= ref_clk;
    sim_rst      <= ref_rst;
  
    ctlr_miso.done   <= '0' , '1' AFTER 1000 ps;
    ctlr_miso.cal_ok <= '0' , '1' AFTER 1000 ps;
  
    -- Delay the control one cycle here so p_mem_access can update its outputs
    -- immediately
    p_clk : PROCESS(sim_clk)
    BEGIN
      IF rising_edge(sim_clk) THEN
        sim_ctlr_mosi <= ctlr_mosi;
      END IF;
    END PROCESS;
  
    p_mem_access : PROCESS(sim_clk, sim_ctlr_mosi)
      VARIABLE v_enable_model  : BOOLEAN := FALSE;
      VARIABLE v_mem_arr       : t_mem_arr := (OTHERS=>(OTHERS=>'0'));
      VARIABLE v_address       : NATURAL := 0;
      VARIABLE v_wr_bursting   : BOOLEAN := FALSE;
      VARIABLE v_rd_bursting   : BOOLEAN := FALSE;
      VARIABLE v_burst_cnt     : NATURAL := 0;
      VARIABLE v_burst_size    : NATURAL := 0;
    BEGIN
  
      -- Don't waste simulation time when user does not use the model anyway
      IF v_enable_model = FALSE THEN
        ctlr_miso.waitrequest_n <= '1'; 
        IF sim_ctlr_mosi.burstbegin='1' THEN
          v_enable_model := TRUE;
        END IF;
      END IF;
  
      IF v_enable_model = TRUE THEN
        IF rising_edge(sim_clk) THEN
    
          ctlr_miso.rdval <= '0'; 
          ctlr_miso.waitrequest_n <= '1'; 
    
          -- Access: burst begin
          IF sim_ctlr_mosi.burstbegin='1' THEN
            IF sim_ctlr_mosi.wr='1' THEN
              v_wr_bursting := TRUE;
            ELSIF sim_ctlr_mosi.rd='1' THEN
              v_rd_bursting := TRUE;
            END IF;
            v_address     := TO_UINT(sim_ctlr_mosi.address);
            v_burst_size  := TO_UINT(sim_ctlr_mosi.burstsize);
            v_burst_cnt   := 0;
          END IF;
      
          IF v_wr_bursting=TRUE OR v_rd_bursting=TRUE THEN
            IF sim_ctlr_mosi.wr='1' THEN -- Write
              v_mem_arr(v_address) := sim_ctlr_mosi.wrdata(c_dat_w-1 DOWNTO 0);
              v_address   := v_address+1;
              v_burst_cnt := v_burst_cnt +1;
            ELSIF v_rd_bursting=TRUE THEN -- Read
              ctlr_miso.rddata(c_dat_w-1 DOWNTO 0) <= v_mem_arr(v_address);  
              ctlr_miso.rdval <= '1';
              ctlr_miso.waitrequest_n <= '0';
              v_address   := v_address+1;
              v_burst_cnt := v_burst_cnt +1;
            END IF;
      
            IF v_burst_cnt = v_burst_size THEN
              v_wr_bursting := FALSE;
              v_rd_bursting := FALSE;
            END IF;
          END IF;
      
          address <= v_address;
    
        END IF;
      END IF;
  
    END PROCESS;
  END GENERATE;
 
END str;

