--------------------------------------------------------------------------------
--
-- Copyright (C) 2011
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- JIVE (Joint Institute for VLBI in Europe) <http://www.jive.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------

-- Purpose: Test bench for the MegaWizard-generated ip_stratixiv_phy_xaui_0.
-- Description:
-- . Block diagram:
--
--                Hard 
--   XGMII  <-->  PCS  <-->  PMA  <--> XAUI serial 
--
-- . With c_sel=0 and c_xgmii_d_test the tb is self checking xgmii_rx_d. 
-- . The link_fault stimulates the link recovery and then also shows that the
--   hi and lo word of the 64b rx_xgmii_d can get swapped. This is not merely
--   an one word offset because with c_xgmii_d_count it shows that the words
--   get swapped.
-- . For the rx_clk it appears that at least in simulation:
--   - rx_clk = tr_clk, as can be checked by using g_tr_clk_156_period = 
--     6.400010 ns.
--   - The rx_clk can be used as tx_clk because the rx_clk does not stop when
--     the link breaks. This is also done in the altera_eth_10g_mac_xaui
--     reference design and also shown in Figure 10.1 in
--     http://www.altera.com/literature/ug/10Gbps_MAC.pdf
--   - The rx_clk phase does not depend on the link delay, so it seems to be
--     from the local tr_clk and not derived from the recovered rx clock.
-- . Status signals:
--   - tx_ready: Indicates PMA TX has exited the reset state and the
--               transceiver can transmit data.
--   - rx_ready: Indicates PMA RX has exited the reset state and the
--               transceiver can receive data.
--   The rx_ready goes high after the tx_ready and both remain high also when
--   the link breaks. In the altera_eth_10g_mac_xaui reference design both are
--   used for tx flow control (ready) and to enable data valid.
--
--   - rx_channelaligned: When asserted, indicates that all 4 RX channels
--                        are aligned.
--   - rx_syncstatus[7:0]: Synchronization indication. RX synchronization is
--                         The indicated on the rx_syncstatus port of each
--                         channel. rx_syncstatus signal is 2 bits per channel
--                         for a total of 8 bits per hard XAUI link.
--   The rx_channelaligned is asserted after the rx_syncstatus, both go low
--   when the link breaks.
-- 
-- Usage:
--   > do wave_tb_ip_stratixiv_phy_xaui.do
--   > run -all

LIBRARY IEEE, common_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_interface_layers_pkg.ALL;

ENTITY tb_ip_stratixiv_phy_xaui IS  
  GENERIC (
    g_sim_time                : TIME := 5 us;
    g_tr_clk_156_period       : TIME := 6.4 ns;
    g_verify_link_recovery    : BOOLEAN := TRUE; 
    g_use_xaui_rx_serial_in   : BOOLEAN := FALSE 
  );
  PORT (
    xaui_tx_serial_out  : OUT STD_LOGIC_VECTOR(c_nof_xaui_lanes-1 DOWNTO 0);
    xaui_rx_serial_in   : IN  STD_LOGIC_VECTOR(c_nof_xaui_lanes-1 DOWNTO 0)
  );
END ENTITY tb_ip_stratixiv_phy_xaui;

ARCHITECTURE str of tb_ip_stratixiv_phy_xaui IS

  CONSTANT c_phy_loopback        : BOOLEAN := NOT g_use_xaui_rx_serial_in;
  CONSTANT c_mm_clk_period       : TIME := 25 ns;   -- 40 MHz

  CONSTANT c_sel                 : NATURAL := 0;
  CONSTANT c_xgmii_d_test        : STD_LOGIC_VECTOR(c_xgmii_data_w-1 DOWNTO 0) := x"DEADBEEFCAFEBABE";  -- c_sel = 0
  CONSTANT c_xgmii_d_count       : STD_LOGIC_VECTOR(c_xgmii_data_w-1 DOWNTO 0) := x"0001020304050607";  -- c_sel = 1
 
  SIGNAL tr_clk                  : STD_LOGIC := '0';
  SIGNAL tx_clk                  : STD_LOGIC := '0';
  SIGNAL rx_clk                  : STD_LOGIC;

  SIGNAL mm_clk                  : STD_LOGIC := '0';
  SIGNAL mm_rst                  : STD_LOGIC := '1';

  SIGNAL link_fault              : STD_LOGIC := '0';
  SIGNAL xaui_tx_serial          : STD_LOGIC_VECTOR(c_nof_xaui_lanes-1 DOWNTO 0);
  SIGNAL xaui_tx_serial_dly      : STD_LOGIC_VECTOR(c_nof_xaui_lanes-1 DOWNTO 0);
  SIGNAL xaui_rx_serial          : STD_LOGIC_VECTOR(c_nof_xaui_lanes-1 DOWNTO 0);

  --XGMII data and control combined:
  SIGNAL xgmii_tx_dc             : STD_LOGIC_VECTOR(c_xgmii_w-1 DOWNTO 0);
  SIGNAL xgmii_rx_dc             : STD_LOGIC_VECTOR(c_xgmii_w-1 DOWNTO 0);

  --XGMII control bits (one for each XGMII lane):
  SIGNAL xgmii_tx_c              : STD_LOGIC_VECTOR(c_xgmii_ctrl_w-1 DOWNTO 0);
  SIGNAL xgmii_rx_c              : STD_LOGIC_VECTOR(c_xgmii_ctrl_w-1 DOWNTO 0);

  --XGMII data
  SIGNAL xgmii_tx_d              : STD_LOGIC_VECTOR(c_xgmii_data_w-1 DOWNTO 0);
  SIGNAL xgmii_rx_d              : STD_LOGIC_VECTOR(c_xgmii_data_w-1 DOWNTO 0);

  -- Important status signal: only send user data after you know the receiver's channel_aligned='1'.
  SIGNAL rx_channelaligned       : STD_LOGIC;

BEGIN

  tr_clk  <= NOT tr_clk AFTER g_tr_clk_156_period/2;    
  tx_clk  <= NOT tx_clk AFTER g_tr_clk_156_period/2;    
  mm_clk  <= NOT mm_clk AFTER c_mm_clk_period/2; 

  p_mm_rst : PROCESS 
  BEGIN
    mm_rst<='1';
--    WAIT FOR 4*c_mm_clk_period; -- Releasing reset here instead of after 80ns will cause the 64bit RX boundary to be 'correct'.
    WAIT FOR 6*c_mm_clk_period; -- We don't want to base rest of development on lucky alignment. This 80ns delay gives us a normal, 32-bit RX boundary.
    WAIT UNTIL rising_edge(mm_clk);
    mm_rst<='0';  
    WAIT;
  END PROCESS;
  
  p_xgmii_data_ctrl : PROCESS
  BEGIN
    WHILE TRUE LOOP
      xgmii_tx_c <= c_xgmii_c_idle;
      xgmii_tx_d <= c_xgmii_d_idle;
  
      WAIT UNTIL rx_channelaligned = '1';
      WAIT UNTIL rising_edge(tx_clk);
  
      xgmii_tx_c <= c_xgmii_c_data;
      IF c_sel=0 THEN
        xgmii_tx_d <= c_xgmii_d_test;
      ELSE
        xgmii_tx_d <= c_xgmii_d_count;
      END IF;
      WAIT UNTIL rising_edge(tx_clk);
      
      WHILE rx_channelaligned = '1' LOOP
        IF c_sel=1 THEN
          FOR I IN 0 TO 7 LOOP
            xgmii_tx_d((I+1)*8-1 DOWNTO I*8) <= INCR_UVEC(xgmii_tx_d((I+1)*8-1 DOWNTO I*8), 8);
          END LOOP;
        END IF;
        WAIT UNTIL rising_edge(tx_clk);
      END LOOP;
    END LOOP;

    WAIT;
  END PROCESS;

  p_xgmii_data_verify : PROCESS
  BEGIN
    IF c_sel=0 THEN
      WAIT FOR 2 us;
      ASSERT xgmii_rx_d = c_xgmii_d_test OR 
             xgmii_rx_d = c_xgmii_d_test(31 DOWNTO 0) & c_xgmii_d_test(63 DOWNTO 32)
             REPORT "tb_ip_stratixiv_phy_xaui : Wrong xgmii_rx_d result value" SEVERITY ERROR;
      WAIT FOR g_sim_time - 2 us;
      ASSERT xgmii_rx_d = c_xgmii_d_test OR 
             xgmii_rx_d = c_xgmii_d_test(31 DOWNTO 0) & c_xgmii_d_test(63 DOWNTO 32)
             REPORT "tb_ip_stratixiv_phy_xaui : Wrong xgmii_rx_d result value" SEVERITY ERROR;
    END IF;
    ASSERT FALSE REPORT "Simulation finished." SEVERITY FAILURE;
    WAIT;
  END PROCESS;
  
  -- Combine data and control into XGMII
  xgmii_tx_dc <= func_xgmii_dc(xgmii_tx_d, xgmii_tx_c);

  -- Extract data (d) from combined data+control (dc) XGMII
  xgmii_rx_d <= func_xgmii_d(xgmii_rx_dc);
  xgmii_rx_c <= func_xgmii_c(xgmii_rx_dc);

  -- DUT:
  u_ip_phy_xaui : ENTITY work.ip_stratixiv_phy_xaui_0
  PORT MAP (
  	pll_ref_clk          => tr_clk,
  	xgmii_tx_clk         => tx_clk,  -- rx_clk
  	xgmii_rx_clk         => rx_clk,
  	xgmii_rx_dc          => xgmii_rx_dc,
  	xgmii_tx_dc          => xgmii_tx_dc,
  	xaui_rx_serial_data  => xaui_rx_serial,
  	xaui_tx_serial_data  => xaui_tx_serial,
  	rx_ready             => OPEN,
  	tx_ready             => OPEN,
  	phy_mgmt_clk         => mm_clk,
  	phy_mgmt_clk_reset   => mm_rst,
  	phy_mgmt_address     => (OTHERS=>'0'),
  	phy_mgmt_read        => '0',
  	phy_mgmt_readdata    => OPEN,
  	phy_mgmt_write       => '0',
  	phy_mgmt_writedata   => (OTHERS=>'0'),
  	phy_mgmt_waitrequest => OPEN,

    rx_channelaligned    => rx_channelaligned
  );
  
  xaui_tx_serial_dly <= TRANSPORT xaui_tx_serial AFTER 100 ps;
  
  gen_verify_link_recovery : IF g_verify_link_recovery=TRUE GENERATE
    link_fault <= '1' AFTER 2.5 us, '0' AFTER 3 us;
  END GENERATE;
  
  xaui_tx_serial_dly <= TRANSPORT xaui_tx_serial AFTER 100 ps;
  xaui_tx_serial_out <= xaui_tx_serial_dly;
  
  p_xaui_rx_serial : PROCESS(xaui_tx_serial_dly, xaui_rx_serial_in, link_fault)
  BEGIN
    IF c_phy_loopback=TRUE THEN
      xaui_rx_serial <= xaui_tx_serial_dly;
    ELSE
      xaui_rx_serial <= xaui_rx_serial_in;
    END IF;
    IF link_fault='1' THEN
      xaui_rx_serial <= (OTHERS=>'0');
    END IF;
  END PROCESS;

END ARCHITECTURE str;

