-------------------------------------------------------------------------------
--
-- Copyright (C) 2014
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

-- Purpose: Reusable test bench verification for the tech_mac_10g
-- Description:
--   Based on the transmitted data at eop verify the received data at eop.

LIBRARY IEEE, technology_lib, common_lib, dp_lib;
USE IEEE.std_logic_1164.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_network_total_header_pkg.ALL;
USE common_lib.tb_common_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;
USE WORK.tb_tech_mac_10g_pkg.ALL;


ENTITY tb_tech_mac_10_verify_rx_at_eop IS
  GENERIC (
    g_no_padding     : BOOLEAN := FALSE;  -- default FALSE to verify the data from MAC that is padded to minimal length, else use TRUE to verify the data of any length
    g_pkt_length_arr : t_nat_natural_arr
  );
  PORT (
    tx_clk      : IN  STD_LOGIC;
    tx_sosi     : IN  t_dp_sosi;
    rx_clk      : IN  STD_LOGIC;
    rx_sosi     : IN  t_dp_sosi
  );
END tb_tech_mac_10_verify_rx_at_eop;

ARCHITECTURE tb OF tb_tech_mac_10_verify_rx_at_eop IS

  CONSTANT c_nof_pkt       : NATURAL := g_pkt_length_arr'LENGTH;
  
  SIGNAL expected_sosi_arr : t_dp_sosi_arr(0 TO c_nof_pkt-1);
  SIGNAL rx_sosi_reg       : t_dp_sosi;
   
BEGIN
  
  p_ff_store_tx_at_eop : PROCESS(tx_clk)
    VARIABLE vI : NATURAL := 0;
  BEGIN
    IF rising_edge(tx_clk) THEN
      IF tx_sosi.eop='1' THEN
        expected_sosi_arr(vI) <= tx_sosi;
        vI := vI +1;
      END IF;
    END IF;
  END PROCESS;

  p_ff_verify_rx_at_eop : PROCESS(rx_clk)
    VARIABLE vI   : NATURAL := 0;
    VARIABLE vLow : NATURAL := 0;
  BEGIN
    IF rising_edge(rx_clk) THEN
      rx_sosi_reg <= rx_sosi;  -- use rx_sosi_reg for verification at eop to account for once cycle latency in expected_sosi_arr()
      IF rx_sosi_reg.eop='1' THEN
        IF g_no_padding=FALSE AND g_pkt_length_arr(vI) < 64 - 14 - 20 - 8 - 4 THEN  -- = minimum frame 64 - ETH 14 - IP 20 - UDP 8 - CRC 4
          -- frame shorter than 64 get padded by MAC so empty after stripping the Rx CRC is fixed 4, which becomes 6 due to pre header padding for UDP word align
          IF TO_UINT(rx_sosi_reg.empty) /= 6 THEN
            REPORT "RX at eop: Wrong padded empty" SEVERITY ERROR;
          END IF;
        ELSE
          IF rx_sosi_reg.empty /= expected_sosi_arr(vI).empty THEN
            REPORT "RX at eop: Wrong empty" SEVERITY ERROR;
          ELSE
            vLow := TO_UINT(rx_sosi_reg.empty)*8;
            ASSERT rx_sosi_reg.data(63 DOWNTO vLow) = expected_sosi_arr(vI).data(63 DOWNTO vLow)  REPORT "RX at eop: Wrong data" SEVERITY ERROR;
          END IF;
        END IF;
        vI := vI +1;
      END IF;
    END IF;
  END PROCESS;

END tb;
