--------------------------------------------------------------------------------
--
-- Copyright (C) 2012
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- JIVE (Joint Institute for VLBI in Europe) <http://www.jive.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------
 
LIBRARY IEEE, common_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE IEEE.NUMERIC_STD.ALL;
USE common_lib.common_pkg.ALL;

ENTITY tech_transceiver_tx_rst IS
  GENERIC(
    g_nof_gx             : NATURAL
    );
  PORT(      
    tr_clk               : IN  STD_LOGIC;
    tr_rst               : IN  STD_LOGIC;
    trc_tx_pll_powerdown : OUT STD_LOGIC;
    tx_pll_locked        : IN  STD_LOGIC_VECTOR(g_nof_gx-1 DOWNTO 0);
    trc_tx_digital_rst   : OUT STD_LOGIC;
    trc_tx_rdy           : OUT STD_LOGIC
   );  
END tech_transceiver_tx_rst;

ARCHITECTURE rtl OF tech_transceiver_tx_rst IS

  -- The minimum pulse width in to keep trc_tx_pll_powerdown asserted = 1us (stratix iv device datasheet).
  -- At 156.25Mhz (6.4ns) this takes 156.25 cycles. As a lower data rate increases the tx_clk period,
  -- we're already on the safe side using this value.
  CONSTANT c_pll_powerdown   : NATURAL := 157; -- 156.25 cycles
  CONSTANT c_pll_powerdown_w : NATURAL := ceil_log2(c_pll_powerdown);

  TYPE t_state_enum IS (s_init, s_power_down_pll, s_wait_for_pll_lock, s_trc_tx_rdy);

  SIGNAL state                    : t_state_enum;
  SIGNAL nxt_state                : t_state_enum; 

  SIGNAL i_trc_tx_pll_powerdown   : STD_LOGIC;
  SIGNAL nxt_trc_tx_pll_powerdown : STD_LOGIC;

  SIGNAL i_trc_tx_digital_rst     : STD_LOGIC;
  SIGNAL nxt_trc_tx_digital_rst   : STD_LOGIC;

  SIGNAL i_trc_tx_rdy             : STD_LOGIC;
  SIGNAL nxt_trc_tx_rdy           : STD_LOGIC;

  SIGNAL cycle_cnt                : STD_LOGIC_VECTOR(c_pll_powerdown_w-1 DOWNTO 0);
  SIGNAL nxt_cycle_cnt            : STD_LOGIC_VECTOR(c_pll_powerdown_w-1 DOWNTO 0);

  SIGNAL trc_tx_pll_locked        : STD_LOGIC_VECTOR(g_nof_gx-1 DOWNTO 0);

BEGIN 

  trc_tx_pll_powerdown <= i_trc_tx_pll_powerdown;
  trc_tx_digital_rst   <= i_trc_tx_digital_rst;
  trc_tx_rdy           <= i_trc_tx_rdy;

  p_clk : PROCESS(tr_rst, tr_clk)
  BEGIN
    IF tr_rst='1' THEN
      state                  <= s_init;  
      i_trc_tx_pll_powerdown <= '1';
      i_trc_tx_digital_rst   <= '1';
      i_trc_tx_rdy           <= '0';
      cycle_cnt              <= (OTHERS=>'0');
    ELSIF rising_edge(tr_clk) THEN
      state                  <= nxt_state;
      i_trc_tx_pll_powerdown <= nxt_trc_tx_pll_powerdown;
      i_trc_tx_digital_rst   <= nxt_trc_tx_digital_rst;
      i_trc_tx_rdy           <= nxt_trc_tx_rdy;
      cycle_cnt              <= nxt_cycle_cnt;
    END IF;
  END PROCESS;

  p_state : PROCESS(state, i_trc_tx_pll_powerdown, i_trc_tx_digital_rst, i_trc_tx_rdy, cycle_cnt, trc_tx_pll_locked)
  BEGIN  
    nxt_state                <= state;  
    nxt_trc_tx_pll_powerdown <= i_trc_tx_pll_powerdown;
    nxt_trc_tx_digital_rst   <= i_trc_tx_digital_rst;
    nxt_trc_tx_rdy           <= i_trc_tx_rdy;
    nxt_cycle_cnt            <= cycle_cnt;

    CASE state IS
      WHEN s_init => 
        nxt_state <= s_power_down_pll;     
    
      WHEN s_power_down_pll => -- Wait until we're sure the PLL is powered down, then power it up
        nxt_cycle_cnt <= INCR_UVEC(cycle_cnt, 1); 
        IF cycle_cnt = TO_UVEC(c_pll_powerdown, c_pll_powerdown_w) THEN
          nxt_trc_tx_pll_powerdown <= '0'; 
          nxt_state                <= s_wait_for_pll_lock;
        END IF;
  
      WHEN s_wait_for_pll_lock => -- Wait until PLL is locked    
        IF andv(trc_tx_pll_locked) = '1' THEN
          nxt_trc_tx_digital_rst <= '0';
          nxt_state              <= s_trc_tx_rdy;
        END IF; 
    
      WHEN s_trc_tx_rdy => -- Transmitter ready to use.
        nxt_trc_tx_rdy <= '1';
  
      WHEN OTHERS =>
        nxt_state <= s_power_down_pll;
    END CASE;
  END PROCESS;

  gen_asyncs: FOR i IN 0 TO g_nof_gx-1 GENERATE
    u_async_pll_locked : ENTITY common_lib.common_async
    GENERIC MAP(
      g_rst_level => '0'
    )
    PORT MAP(
      rst  => tr_rst,
      clk  => tr_clk,
      din  => tx_pll_locked(i),
      dout => trc_tx_pll_locked(i)
    );
  END GENERATE;

END rtl;

